# Generated by Django 4.1.3 on 2022-12-27 23:22

import re
import uuid
from decimal import Decimal

import django.core.validators
import django.db.models.manager
from django.conf import settings
from django.db import migrations, models

import django_ledger.io.io_core
import django_ledger.models.entity
import django_ledger.models.mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AccountModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('path', models.CharField(max_length=255, unique=True)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('code', models.CharField(max_length=10, verbose_name='Account Code')),
                ('name', models.CharField(max_length=100, verbose_name='Account Name')),
                ('role', models.CharField(choices=[('Assets', (
                ('asset_ca_cash', 'Current Asset'), ('asset_ca_mkt_sec', 'Marketable Securities'),
                ('asset_ca_recv', 'Receivables'), ('asset_ca_inv', 'Inventory'), ('asset_ca_uncoll', 'Uncollectibles'),
                ('asset_ca_prepaid', 'Prepaid'), ('asset_ca_other', 'Other Liquid Assets'),
                ('asset_lti_notes', 'Notes Receivable'), ('asset_lti_land', 'Land'), ('asset_lti_sec', 'Securities'),
                ('asset_ppe_build', 'Buildings'), ('asset_ppe_build_accum_depr', 'Buildings - Accum. Depreciation'),
                ('asset_ppe_plant', 'Plant'), ('asset_ppe_plant_depr', 'Plant - Accum. Depreciation'),
                ('asset_ppe_equip', 'Equipment'), ('asset_ppe_equip_accum_depr', 'Equipment - Accum. Depreciation'),
                ('asset_ia', 'Intangible Assets'), ('asset_ia_accum_amort', 'Intangible Assets - Accum. Amortization'),
                ('asset_adjustment', 'Other Assets'))), ('Liabilities', (
                ('lia_cl_acc_payable', 'Accounts Payable'), ('lia_cl_wages_payable', 'Wages Payable'),
                ('lia_cl_int_payable', 'Interest Payable'), ('lia_cl_taxes_payable', 'Taxes Payable'),
                ('lia_cl_st_notes_payable', 'Notes Payable'),
                ('lia_cl_ltd_mat', 'Current Maturities of Long Term Debt'), ('lia_cl_def_rev', 'Deferred Revenue'),
                ('lia_cl_other', 'Other Liabilities'), ('lia_ltl_notes', 'Notes Payable'),
                ('lia_ltl_bonds', 'Bonds Payable'), ('lia_ltl_mortgage', 'Mortgage Payable'))), ('Equity', (
                ('eq_capital', 'Capital'), ('eq_stock_common', 'Common Stock'),
                ('eq_stock_preferred', 'Preferred Stock'), ('eq_adjustment', 'Other Equity Adjustments'),
                ('eq_dividends', 'Dividends & Distributions to Shareholders'), ('in_operational', 'Operational Income'),
                ('in_passive', 'Investing/Passive Income'), ('in_interest', 'Interest Income'),
                ('in_gain_loss', 'Capital Gain/Loss Income'), ('in_other', 'Other Income'),
                ('ex_cogs', 'Cost of Goods Sold'), ('ex_regular', 'Regular Expense'),
                ('ex_interest', 'Interest Expense'), ('ex_taxes', 'Tax Expense'), ('ex_capital', 'Capital Expense'),
                ('ex_depreciation', 'Depreciation Expense'), ('ex_amortization', 'Amortization Expense'),
                ('ex_other', 'Other Expense')))], max_length=30, verbose_name='Account Role')),
                ('balance_type', models.CharField(choices=[('credit', 'Credit'), ('debit', 'Debit')], max_length=6,
                                                  verbose_name='Account Balance Type')),
                ('locked', models.BooleanField(default=False, verbose_name='Locked')),
                ('active', models.BooleanField(default=False, verbose_name='Active')),
            ],
            options={
                'verbose_name': 'Account',
                'verbose_name_plural': 'Accounts',
                'ordering': ['-created'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BillModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('amount_due',
                 models.DecimalField(decimal_places=2, default=0, max_digits=20, verbose_name='Amount Due')),
                ('amount_paid', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Paid')),
                ('amount_receivable', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Receivable')),
                ('amount_unearned', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Unearned')),
                ('amount_earned', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Earned')),
                ('accrue', models.BooleanField(default=False, verbose_name='Accrue')),
                ('progress', models.DecimalField(decimal_places=2, default=0, max_digits=3,
                                                 validators=[django.core.validators.MinValueValidator(limit_value=0),
                                                             django.core.validators.MaxValueValidator(limit_value=1)],
                                                 verbose_name='Progress Amount')),
                ('terms', models.CharField(
                    choices=[('on_receipt', 'Due On Receipt'), ('net_30', 'Net 30 Days'), ('net_60', 'Net 60 Days'),
                             ('net_90', 'Net 90 Days')], default='on_receipt', max_length=10, verbose_name='Terms')),
                ('date_due', models.DateField(blank=True, null=True, verbose_name='Due Date')),
                ('markdown_notes', models.TextField(blank=True, null=True, verbose_name='Markdown Notes')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('bill_number', models.SlugField(editable=False, max_length=20, verbose_name='Bill Number')),
                ('bill_status', models.CharField(
                    choices=[('draft', 'Draft'), ('in_review', 'In Review'), ('approved', 'Approved'), ('paid', 'Paid'),
                             ('canceled', 'Canceled'), ('void', 'Void')], default='draft', max_length=10,
                    verbose_name='Bill Status')),
                ('xref', models.SlugField(blank=True, null=True, verbose_name='External Reference Number')),
                ('additional_info',
                 models.JSONField(blank=True, default=dict, null=True, verbose_name='Bill Additional Info')),
                ('date_draft', models.DateField(blank=True, null=True, verbose_name='Draft Date')),
                ('date_in_review', models.DateField(blank=True, null=True, verbose_name='In Review Date')),
                ('date_approved', models.DateField(blank=True, null=True, verbose_name='Approved Date')),
                ('date_paid', models.DateField(blank=True, null=True, verbose_name='Paid Date')),
                ('date_void', models.DateField(blank=True, null=True, verbose_name='Void Date')),
                ('date_canceled', models.DateField(blank=True, null=True, verbose_name='Canceled Date')),
            ],
            options={
                'verbose_name': 'Bill',
                'verbose_name_plural': 'Bills',
                'ordering': ['-updated'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ChartOfAccountModel',
            fields=[
                ('slug', models.SlugField(editable=False, unique=True, validators=[
                    django.core.validators.MinLengthValidator(limit_value=10,
                                                              message='Slug field must contain at least 10 characters.')])),
                ('name', models.CharField(blank=True, max_length=150, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('locked', models.BooleanField(default=False, verbose_name='Locked')),
                ('description', models.TextField(blank=True, null=True, verbose_name='CoA Description')),
            ],
            options={
                'verbose_name': 'Chart of Account',
                'verbose_name_plural': 'Chart of Accounts',
                'ordering': ['-created'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CustomerModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('address_1', models.CharField(max_length=70, verbose_name='Address Line 1')),
                ('address_2', models.CharField(blank=True, max_length=70, null=True, verbose_name='Address Line 2')),
                ('city', models.CharField(blank=True, max_length=70, null=True, verbose_name='City')),
                ('state', models.CharField(blank=True, max_length=70, null=True, verbose_name='State/Province')),
                ('zip_code', models.CharField(blank=True, max_length=20, null=True, verbose_name='Zip Code')),
                ('country', models.CharField(blank=True, max_length=70, null=True, verbose_name='Country')),
                ('email', models.EmailField(blank=True, max_length=254, null=True, verbose_name='Email')),
                ('website', models.URLField(blank=True, null=True, verbose_name='Website')),
                ('phone', models.CharField(blank=True, max_length=30, null=True, verbose_name='Phone Number')),
                ('sales_tax_rate', models.FloatField(blank=True, default=0.0, null=True, validators=[
                    django.core.validators.MinValueValidator(limit_value=0.0),
                    django.core.validators.MaxValueValidator(limit_value=1.0)], verbose_name='Sales Tax Rate')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('customer_name', models.CharField(max_length=100)),
                ('customer_number', models.CharField(editable=False, max_length=30, verbose_name='Customer Number')),
                ('description', models.TextField()),
                ('active', models.BooleanField(default=True)),
                ('hidden', models.BooleanField(default=False)),
                ('additional_info', models.JSONField(blank=True, default=dict, null=True)),
            ],
            options={
                'verbose_name': 'Customer',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EntityManagementModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('permission_level', models.CharField(
                    choices=[('read', 'Read Permissions'), ('write', 'Read/Write Permissions'),
                             ('suspended', 'No Permissions')], default='read', max_length=10,
                    verbose_name='Permission Level')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EntityModel',
            fields=[
                ('slug', models.SlugField(editable=False, unique=True, validators=[
                    django.core.validators.MinLengthValidator(limit_value=10,
                                                              message='Slug field must contain at least 10 characters.')])),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('address_1', models.CharField(max_length=70, verbose_name='Address Line 1')),
                ('address_2', models.CharField(blank=True, max_length=70, null=True, verbose_name='Address Line 2')),
                ('city', models.CharField(blank=True, max_length=70, null=True, verbose_name='City')),
                ('state', models.CharField(blank=True, max_length=70, null=True, verbose_name='State/Province')),
                ('zip_code', models.CharField(blank=True, max_length=20, null=True, verbose_name='Zip Code')),
                ('country', models.CharField(blank=True, max_length=70, null=True, verbose_name='Country')),
                ('email', models.EmailField(blank=True, max_length=254, null=True, verbose_name='Email')),
                ('website', models.URLField(blank=True, null=True, verbose_name='Website')),
                ('phone', models.CharField(blank=True, max_length=30, null=True, verbose_name='Phone Number')),
                ('path', models.CharField(max_length=255, unique=True)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=150, verbose_name='Entity Name')),
                ('hidden', models.BooleanField(default=False)),
                ('accrual_method', models.BooleanField(default=False, verbose_name='Use Accrual Method')),
                ('fy_start_month', models.IntegerField(
                    choices=[(1, 'January'), (2, 'February'), (3, 'March'), (4, 'April'), (5, 'May'), (6, 'June'),
                             (7, 'July'), (8, 'August'), (9, 'September'), (10, 'October'), (11, 'November'),
                             (12, 'December')], default=1, verbose_name='Fiscal Year Start')),
                ('picture', models.ImageField(blank=True, null=True, upload_to='')),
                ('admin', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='admin_of',
                                            to=settings.AUTH_USER_MODEL, verbose_name='Admin')),
                ('default_coa', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT,
                                                     to='django_ledger.chartofaccountmodel',
                                                     verbose_name='Default Chart of Accounts')),
                ('managers',
                 models.ManyToManyField(related_name='managed_by', through='django_ledger.EntityManagementModel',
                                        to=settings.AUTH_USER_MODEL, verbose_name='Managers')),
            ],
            options={
                'verbose_name': 'Entity',
                'verbose_name_plural': 'Entities',
                'ordering': ['-created'],
                'abstract': False,
            },
            bases=(models.Model, django_ledger.io.io_core.IOMixIn, django_ledger.models.mixins.LoggingMixIn,
                   django_ledger.models.entity.EntityModelFiscalPeriodMixIn),
        ),
        migrations.CreateModel(
            name='EntityUnitModel',
            fields=[
                ('name', models.CharField(blank=True, max_length=150, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('path', models.CharField(max_length=255, unique=True)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('slug', models.SlugField()),
                ('document_prefix', models.CharField(max_length=3)),
                ('active', models.BooleanField(default=True, verbose_name='Is Active')),
                ('hidden', models.BooleanField(default=False, verbose_name='Is Hidden')),
                ('entity', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                             to='django_ledger.entitymodel', verbose_name='Unit Entity')),
            ],
            options={
                'verbose_name': 'Entity Unit Model',
                'ordering': ['-created'],
                'abstract': False,
            },
            bases=(django_ledger.io.io_core.IOMixIn, models.Model),
        ),
        migrations.CreateModel(
            name='EstimateModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('markdown_notes', models.TextField(blank=True, null=True, verbose_name='Markdown Notes')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('estimate_number', models.SlugField(editable=False, max_length=20, verbose_name='Estimate Number')),
                ('terms', models.CharField(
                    choices=[('fixed', 'Fixed Price'), ('target', 'Target Price'), ('t&m', 'Time & Materials'),
                             ('other', 'Other')], max_length=10, verbose_name='Contract Terms')),
                ('title', models.CharField(max_length=250, validators=[
                    django.core.validators.MinLengthValidator(limit_value=5,
                                                              message='PO Title length must be greater than 5')],
                                           verbose_name='Customer Estimate Title')),
                ('status', models.CharField(
                    choices=[('draft', 'Draft'), ('in_review', 'In Review'), ('approved', 'Approved'),
                             ('completed', 'Completed'), ('void', 'Void'), ('canceled', 'Canceled')], default='draft',
                    max_length=10, verbose_name='Customer Estimate Status')),
                ('date_draft', models.DateField(blank=True, null=True, verbose_name='Date Draft')),
                ('date_in_review', models.DateField(blank=True, null=True, verbose_name='Date In Review')),
                ('date_approved', models.DateField(blank=True, null=True, verbose_name='Date Approved')),
                ('date_completed', models.DateField(blank=True, null=True, verbose_name='Date Completed')),
                ('date_canceled', models.DateField(blank=True, null=True, verbose_name='Date Canceled')),
                ('date_void', models.DateField(blank=True, null=True, verbose_name='Date Void')),
                ('revenue_estimate', models.DecimalField(decimal_places=2, default=Decimal('0.00'),
                                                         help_text='Estimated cost to complete the quoted work.',
                                                         max_digits=20,
                                                         validators=[django.core.validators.MinValueValidator(0)],
                                                         verbose_name='Total revenue estimate')),
                ('labor_estimate', models.DecimalField(decimal_places=2, default=Decimal('0.00'),
                                                       help_text='Estimated labor cost to complete the quoted work.',
                                                       max_digits=20,
                                                       validators=[django.core.validators.MinValueValidator(0)],
                                                       verbose_name='Labor Cost of labor estimate')),
                ('material_estimate', models.DecimalField(decimal_places=2, default=0.0,
                                                          help_text='Estimated material cost to complete the quoted work.',
                                                          max_digits=20,
                                                          validators=[django.core.validators.MinValueValidator(0)],
                                                          verbose_name='Material Cost Estimate')),
                ('equipment_estimate', models.DecimalField(decimal_places=2, default=Decimal('0.00'),
                                                           help_text='Estimated equipment cost to complete the quoted work.',
                                                           max_digits=20,
                                                           validators=[django.core.validators.MinValueValidator(0)],
                                                           verbose_name='Equipment Cost Estimate')),
                ('other_estimate', models.DecimalField(decimal_places=2, default=Decimal('0.00'),
                                                       help_text='Estimated equipment cost to complete the quoted work.',
                                                       max_digits=20,
                                                       validators=[django.core.validators.MinValueValidator(0)],
                                                       verbose_name='Other Cost Estimate')),
                ('customer',
                 models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.customermodel',
                                   verbose_name='Customer')),
                ('entity', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                             to='django_ledger.entitymodel', verbose_name='Entity Model')),
            ],
            options={
                'verbose_name': 'Customer Job',
                'verbose_name_plural': 'Customer Jobs',
                'ordering': ['-updated'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ImportJobModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('description', models.CharField(max_length=200, verbose_name='Description')),
                ('completed', models.BooleanField(default=False, verbose_name='Import Job Completed')),
            ],
            options={
                'verbose_name': 'Import Job Model',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='InvoiceModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('amount_due',
                 models.DecimalField(decimal_places=2, default=0, max_digits=20, verbose_name='Amount Due')),
                ('amount_paid', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Paid')),
                ('amount_receivable', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Receivable')),
                ('amount_unearned', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Unearned')),
                ('amount_earned', models.DecimalField(decimal_places=2, default=0, max_digits=20, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Amount Earned')),
                ('accrue', models.BooleanField(default=False, verbose_name='Accrue')),
                ('progress', models.DecimalField(decimal_places=2, default=0, max_digits=3,
                                                 validators=[django.core.validators.MinValueValidator(limit_value=0),
                                                             django.core.validators.MaxValueValidator(limit_value=1)],
                                                 verbose_name='Progress Amount')),
                ('terms', models.CharField(
                    choices=[('on_receipt', 'Due On Receipt'), ('net_30', 'Net 30 Days'), ('net_60', 'Net 60 Days'),
                             ('net_90', 'Net 90 Days')], default='on_receipt', max_length=10, verbose_name='Terms')),
                ('date_due', models.DateField(blank=True, null=True, verbose_name='Due Date')),
                ('markdown_notes', models.TextField(blank=True, null=True, verbose_name='Markdown Notes')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('invoice_number', models.SlugField(editable=False, max_length=20, verbose_name='Invoice Number')),
                ('invoice_status', models.CharField(
                    choices=[('draft', 'Draft'), ('in_review', 'In Review'), ('approved', 'Approved'), ('paid', 'Paid'),
                             ('void', 'Void'), ('canceled', 'Canceled')], default='draft', max_length=10,
                    verbose_name='Invoice Status')),
                ('additional_info',
                 models.JSONField(blank=True, default=dict, null=True, verbose_name='Invoice Additional Info')),
                ('date_draft', models.DateField(blank=True, null=True, verbose_name='Draft Date')),
                ('date_in_review', models.DateField(blank=True, null=True, verbose_name='In Review Date')),
                ('date_approved', models.DateField(blank=True, null=True, verbose_name='Approved Date')),
                ('date_paid', models.DateField(blank=True, null=True, verbose_name='Paid Date')),
                ('date_void', models.DateField(blank=True, null=True, verbose_name='Void Date')),
                ('date_canceled', models.DateField(blank=True, null=True, verbose_name='Canceled Date')),
                ('cash_account',
                 models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, related_name='invoice_cash_account',
                                   to='django_ledger.accountmodel', verbose_name='Cash Account')),
                ('ce_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                               to='django_ledger.estimatemodel',
                                               verbose_name='Associated Customer Job/Estimate')),
                ('customer',
                 models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.customermodel',
                                   verbose_name='Customer')),
            ],
            options={
                'verbose_name': 'Invoice',
                'verbose_name_plural': 'Invoices',
                'ordering': ['-updated'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ItemModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('path', models.CharField(max_length=255, unique=True)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=100, verbose_name='Item Name')),
                ('item_type', models.CharField(blank=True,
                                               choices=[('L', 'Labor'), ('M', 'Material'), ('E', 'Equipment'),
                                                        ('S', 'Lump Sum'), ('O', 'Other')], max_length=1, null=True)),
                ('sku', models.CharField(blank=True, max_length=50, null=True, verbose_name='SKU Code')),
                ('upc', models.CharField(blank=True, max_length=50, null=True, verbose_name='UPC Code')),
                ('item_id', models.CharField(blank=True, max_length=50, null=True, verbose_name='Internal ID')),
                ('item_number', models.CharField(editable=False, max_length=30, verbose_name='Item Number')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is Active')),
                ('default_amount', models.DecimalField(decimal_places=2, default=0, max_digits=20,
                                                       validators=[django.core.validators.MinValueValidator(0)],
                                                       verbose_name='Default monetary value per unit of measure')),
                ('for_inventory', models.BooleanField(help_text='It is an item you require for your inventory.',
                                                      verbose_name='Is an item for inventory')),
                ('is_product_or_service',
                 models.BooleanField(help_text='Is a product or service you sell or provide to customers.',
                                     verbose_name='Is a product or service.')),
                ('sold_as_unit', models.BooleanField(default=False)),
                ('inventory_received', models.DecimalField(blank=True, decimal_places=3, max_digits=20, null=True,
                                                           verbose_name='Total inventory received.')),
                ('inventory_received_value', models.DecimalField(blank=True, decimal_places=2, max_digits=20, null=True,
                                                                 verbose_name='Total value of inventory received.')),
                ('additional_info',
                 models.JSONField(blank=True, default=dict, null=True, verbose_name='Item Additional Info')),
                ('cogs_account', models.ForeignKey(blank=True,
                                                   help_text='COGS account where cost will be recognized on Income Statement.',
                                                   null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                   related_name='item_cogs_account', to='django_ledger.accountmodel',
                                                   verbose_name='COGS Account')),
                ('earnings_account', models.ForeignKey(blank=True,
                                                       help_text='Earnings account where revenue will be recognized on Income Statement.',
                                                       null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                       related_name='item_earnings_account',
                                                       to='django_ledger.accountmodel',
                                                       verbose_name='Earnings Account')),
                ('entity', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                             to='django_ledger.entitymodel', verbose_name='Item Entity')),
                ('expense_account', models.ForeignKey(blank=True,
                                                      help_text='Expense account where cost will be recognized on Income Statement.',
                                                      null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                      related_name='item_expense_account',
                                                      to='django_ledger.accountmodel', verbose_name='Expense Account')),
                ('inventory_account',
                 models.ForeignKey(blank=True, help_text='Inventory account where cost will be capitalized.', null=True,
                                   on_delete=django.db.models.deletion.RESTRICT, related_name='item_inventory_account',
                                   to='django_ledger.accountmodel', verbose_name='Inventory Account')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ItemTransactionModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('quantity', models.FloatField(blank=True, null=True,
                                               validators=[django.core.validators.MinValueValidator(limit_value=0.0)],
                                               verbose_name='Quantity')),
                ('unit_cost', models.FloatField(blank=True, null=True,
                                                validators=[django.core.validators.MinValueValidator(limit_value=0.0)],
                                                verbose_name='Cost Per Unit')),
                ('total_amount',
                 models.DecimalField(blank=True, decimal_places=2, editable=False, max_digits=20, null=True,
                                     validators=[django.core.validators.MinValueValidator(limit_value=0.0)],
                                     verbose_name='Total Amount QTY x UnitCost')),
                ('po_quantity',
                 models.FloatField(blank=True, help_text='Authorized item quantity for purchasing.', null=True,
                                   validators=[django.core.validators.MinValueValidator(limit_value=0.0)],
                                   verbose_name='PO Quantity')),
                ('po_unit_cost', models.FloatField(blank=True, help_text='Purchase Order unit cost.', null=True,
                                                   validators=[
                                                       django.core.validators.MinValueValidator(limit_value=0.0)],
                                                   verbose_name='PO Unit Cost')),
                ('po_total_amount', models.DecimalField(blank=True, decimal_places=2, editable=False,
                                                        help_text='Maximum authorized cost per Purchase Order.',
                                                        max_digits=20, null=True, validators=[
                        django.core.validators.MinValueValidator(limit_value=0.0)],
                                                        verbose_name='Authorized maximum item cost per Purchase Order')),
                ('po_item_status', models.CharField(blank=True,
                                                    choices=[('not_ordered', 'Not Ordered'), ('ordered', 'Ordered'),
                                                             ('in_transit', 'In Transit'), ('received', 'Received'),
                                                             ('cancelled', 'Canceled')], max_length=15, null=True,
                                                    verbose_name='PO Item Status')),
                ('ce_quantity', models.FloatField(blank=True, null=True, validators=[
                    django.core.validators.MinValueValidator(limit_value=0.0)],
                                                  verbose_name='Estimated/Contract Quantity')),
                ('ce_unit_cost_estimate', models.FloatField(blank=True, null=True, validators=[
                    django.core.validators.MinValueValidator(limit_value=0.0)],
                                                            verbose_name='Estimate/Contract Cost per Unit.')),
                ('ce_cost_estimate',
                 models.DecimalField(blank=True, decimal_places=2, editable=False, max_digits=20, null=True,
                                     validators=[django.core.validators.MinValueValidator(limit_value=0.0)],
                                     verbose_name='Total Estimate/Contract Cost.')),
                ('ce_unit_revenue_estimate', models.FloatField(blank=True, null=True, validators=[
                    django.core.validators.MinValueValidator(limit_value=0.0)],
                                                               verbose_name='Estimate/Contract Revenue per Unit.')),
                ('ce_revenue_estimate',
                 models.DecimalField(blank=True, decimal_places=2, editable=False, max_digits=20, null=True,
                                     validators=[django.core.validators.MinValueValidator(limit_value=0.0)],
                                     verbose_name='Total Estimate/Contract Revenue.')),
                ('item_notes', models.CharField(blank=True, max_length=400, null=True, verbose_name='Description')),
                ('bill_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                 to='django_ledger.billmodel', verbose_name='Bill Model')),
                ('ce_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                               to='django_ledger.estimatemodel', verbose_name='Customer Estimate')),
                ('entity_unit', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                  to='django_ledger.entityunitmodel',
                                                  verbose_name='Associated Entity Unit')),
                ('invoice_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                    to='django_ledger.invoicemodel', verbose_name='Invoice Model')),
                ('item_model',
                 models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.itemmodel',
                                   verbose_name='Item Model')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JournalEntryModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('je_number', models.SlugField(editable=False, max_length=20, verbose_name='Journal Entry Number')),
                ('timestamp', models.DateTimeField(verbose_name='Date')),
                ('description', models.CharField(blank=True, max_length=70, null=True, verbose_name='Description')),
                ('activity', models.CharField(blank=True, choices=[('Operating', (('op', 'Operating'),)), ('Investing',
                                                                                                           (('inv_ppe',
                                                                                                             'Purchase/Disposition of PPE'),
                                                                                                            (
                                                                                                            'inv_securities',
                                                                                                            'Purchase/Disposition of Securities'),
                                                                                                            ('inv',
                                                                                                             'Investing Activity Other'))),
                                                                   ('Financing', (
                                                                   ('fin_std', 'Payoff of Short Term Debt'),
                                                                   ('fin_ltd', 'Payoff of Long Term Debt'), (
                                                                   'fin_equity',
                                                                   'Issuance of Common Stock, Preferred Stock or Capital Contribution'),
                                                                   ('fin_dividends',
                                                                    'Dividends or Distributions to Shareholders'),
                                                                   ('fin', 'Financing Activity Other')))],
                                              editable=False, max_length=20, null=True, verbose_name='Activity')),
                ('origin', models.CharField(blank=True, max_length=30, null=True, verbose_name='Origin')),
                ('posted', models.BooleanField(default=False, verbose_name='Posted')),
                ('locked', models.BooleanField(default=False, verbose_name='Locked')),
                ('entity_unit', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                  to='django_ledger.entityunitmodel',
                                                  verbose_name='Associated Entity Unit')),
            ],
            options={
                'verbose_name': 'Journal Entry',
                'verbose_name_plural': 'Journal Entries',
                'ordering': ['-created'],
                'abstract': False,
            },
            managers=[
                ('on_coa', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='VendorModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('address_1', models.CharField(max_length=70, verbose_name='Address Line 1')),
                ('address_2', models.CharField(blank=True, max_length=70, null=True, verbose_name='Address Line 2')),
                ('city', models.CharField(blank=True, max_length=70, null=True, verbose_name='City')),
                ('state', models.CharField(blank=True, max_length=70, null=True, verbose_name='State/Province')),
                ('zip_code', models.CharField(blank=True, max_length=20, null=True, verbose_name='Zip Code')),
                ('country', models.CharField(blank=True, max_length=70, null=True, verbose_name='Country')),
                ('email', models.EmailField(blank=True, max_length=254, null=True, verbose_name='Email')),
                ('website', models.URLField(blank=True, null=True, verbose_name='Website')),
                ('phone', models.CharField(blank=True, max_length=30, null=True, verbose_name='Phone Number')),
                ('account_number', models.CharField(blank=True, max_length=30, null=True, validators=[
                    django.core.validators.RegexValidator(re.compile('^\\d+(?:\\d+)*\\Z'), code='invalid',
                                                          message='Only digits allowed')],
                                                    verbose_name='Account Number')),
                ('routing_number', models.CharField(blank=True, max_length=30, null=True, validators=[
                    django.core.validators.RegexValidator(re.compile('^\\d+(?:\\d+)*\\Z'), code='invalid',
                                                          message='Only digits allowed')],
                                                    verbose_name='Routing Number')),
                ('aba_number', models.CharField(blank=True, max_length=30, null=True, verbose_name='ABA Number')),
                ('swift_number', models.CharField(blank=True, max_length=30, null=True, verbose_name='SWIFT Number')),
                ('account_type',
                 models.CharField(choices=[('checking', 'Checking'), ('savings', 'Savings')], default='checking',
                                  max_length=10, verbose_name='Account Type')),
                ('tax_id_number',
                 models.CharField(blank=True, max_length=30, null=True, verbose_name='Tax Registration Number')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('vendor_number', models.CharField(blank=True, max_length=30, null=True)),
                ('vendor_name', models.CharField(max_length=100)),
                ('description', models.TextField()),
                ('active', models.BooleanField(default=True)),
                ('hidden', models.BooleanField(default=False)),
                ('additional_info', models.JSONField(blank=True, default=dict, null=True)),
                ('entity_model',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel',
                                   verbose_name='Vendor Entity')),
            ],
            options={
                'verbose_name': 'Vendor',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UnitOfMeasureModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=50, verbose_name='Unit of Measure Name')),
                ('unit_abbr', models.SlugField(max_length=10, verbose_name='UoM Abbreviation')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is Active')),
                ('entity', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                             to='django_ledger.entitymodel', verbose_name='UoM Entity')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TransactionModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('tx_type', models.CharField(choices=[('credit', 'Credit'), ('debit', 'Debit')], max_length=10,
                                             verbose_name='Tx Type')),
                ('amount', models.DecimalField(blank=True, decimal_places=2, help_text='Account of the transaction.',
                                               max_digits=20, null=True,
                                               validators=[django.core.validators.MinValueValidator(0)],
                                               verbose_name='Amount')),
                ('description',
                 models.CharField(blank=True, help_text='A description to be included with this individual transaction',
                                  max_length=100, null=True, verbose_name='Tx Description')),
                ('account',
                 models.ForeignKey(help_text='Account from Chart of Accounts to be associated with this transaction.',
                                   on_delete=django.db.models.deletion.PROTECT, to='django_ledger.accountmodel',
                                   verbose_name='Account')),
                ('journal_entry',
                 models.ForeignKey(editable=False, help_text='Journal Entry to be associated with this transaction.',
                                   on_delete=django.db.models.deletion.PROTECT, to='django_ledger.journalentrymodel',
                                   verbose_name='Journal Entry')),
            ],
            options={
                'verbose_name': 'Transaction',
                'verbose_name_plural': 'Transactions',
                'ordering': ['-created'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StagedTransactionModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('fitid', models.CharField(max_length=100)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=15)),
                ('date_posted', models.DateField()),
                ('name', models.CharField(blank=True, max_length=200, null=True)),
                ('memo', models.CharField(blank=True, max_length=200, null=True)),
                ('earnings_account',
                 models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                   to='django_ledger.accountmodel')),
                ('import_job',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.importjobmodel')),
                ('tx', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                            to='django_ledger.transactionmodel')),
            ],
            options={
                'verbose_name': 'Staged Transaction Model',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PurchaseOrderModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('markdown_notes', models.TextField(blank=True, null=True, verbose_name='Markdown Notes')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('po_number', models.SlugField(editable=False, max_length=20, verbose_name='Purchase Order Number')),
                ('po_title', models.CharField(max_length=250, validators=[
                    django.core.validators.MinLengthValidator(limit_value=5,
                                                              message='PO Title must be greater than 5')],
                                              verbose_name='Purchase Order Title')),
                ('po_status', models.CharField(
                    choices=[('draft', 'Draft'), ('in_review', 'In Review'), ('approved', 'Approved'),
                             ('fulfilled', 'Fulfilled'), ('canceled', 'Canceled'), ('void', 'Void')], default='draft',
                    max_length=10)),
                ('po_amount',
                 models.DecimalField(decimal_places=2, default=0, max_digits=20, verbose_name='Purchase Order Amount')),
                ('po_amount_received',
                 models.DecimalField(decimal_places=2, default=0, max_digits=20, verbose_name='Received Amount')),
                ('date_draft', models.DateField(blank=True, null=True, verbose_name='Draft Date')),
                ('date_in_review', models.DateField(blank=True, null=True, verbose_name='In Review Date')),
                ('date_approved', models.DateField(blank=True, null=True, verbose_name='Approved Date')),
                ('date_void', models.DateField(blank=True, null=True, verbose_name='Void Date')),
                ('date_fulfilled', models.DateField(blank=True, null=True, verbose_name='Fulfillment Date')),
                ('date_canceled', models.DateField(blank=True, null=True, verbose_name='Canceled Date')),
                ('ce_model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                               to='django_ledger.estimatemodel',
                                               verbose_name='Associated Customer Job/Estimate')),
                ('entity',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel',
                                   verbose_name='Entity')),
                ('po_items',
                 models.ManyToManyField(through='django_ledger.ItemTransactionModel', to='django_ledger.itemmodel',
                                        verbose_name='Purchase Order Items')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='LedgerModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, max_length=150, null=True, verbose_name='Ledger Name')),
                ('posted', models.BooleanField(default=False, verbose_name='Posted Ledger')),
                ('locked', models.BooleanField(default=False, verbose_name='Locked Ledger')),
                ('hidden', models.BooleanField(default=False, verbose_name='Hidden Ledger')),
                ('entity', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                             to='django_ledger.entitymodel', verbose_name='Ledger Entity')),
            ],
            options={
                'verbose_name': 'Ledger',
                'verbose_name_plural': 'Ledgers',
                'ordering': ['-created'],
                'abstract': False,
            },
            bases=(models.Model, django_ledger.io.io_core.IOMixIn),
        ),
        migrations.AddField(
            model_name='journalentrymodel',
            name='ledger',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='journal_entries',
                                    to='django_ledger.ledgermodel', verbose_name='Ledger'),
        ),
        migrations.AddField(
            model_name='itemtransactionmodel',
            name='po_model',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    to='django_ledger.purchaseordermodel', verbose_name='Purchase Order Model'),
        ),
        migrations.AddField(
            model_name='itemmodel',
            name='uom',
            field=models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, to='django_ledger.unitofmeasuremodel',
                                    verbose_name='Unit of Measure'),
        ),
        migrations.AddField(
            model_name='invoicemodel',
            name='invoice_items',
            field=models.ManyToManyField(through='django_ledger.ItemTransactionModel', to='django_ledger.itemmodel',
                                         verbose_name='Invoice Items'),
        ),
        migrations.AddField(
            model_name='invoicemodel',
            name='ledger',
            field=models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                       to='django_ledger.ledgermodel', verbose_name='Ledger'),
        ),
        migrations.AddField(
            model_name='invoicemodel',
            name='prepaid_account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='invoice_prepaid_account', to='django_ledger.accountmodel',
                                    verbose_name='Prepaid Account'),
        ),
        migrations.AddField(
            model_name='invoicemodel',
            name='unearned_account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='invoice_unearned_account', to='django_ledger.accountmodel',
                                    verbose_name='Unearned Account'),
        ),
        migrations.AddField(
            model_name='importjobmodel',
            name='ledger',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                    to='django_ledger.ledgermodel', verbose_name='Ledger'),
        ),
        migrations.CreateModel(
            name='EntityStateModel',
            fields=[
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('fiscal_year', models.SmallIntegerField(blank=True, null=True, validators=[
                    django.core.validators.MinValueValidator(limit_value=1900)], verbose_name='Fiscal Year')),
                ('key', models.CharField(choices=[('je', 'Journal Entry'), ('po', 'Purchase Order'), ('bill', 'Bill'),
                                                  ('invoice', 'Invoice'), ('estimate', 'Estimate')], max_length=10)),
                ('sequence', models.BigIntegerField(default=0, validators=[
                    django.core.validators.MinValueValidator(limit_value=0)])),
                ('entity_model',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel',
                                   verbose_name='Entity Model')),
                ('entity_unit', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                                  to='django_ledger.entityunitmodel', verbose_name='Entity Unit')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='entitymanagementmodel',
            name='entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entity_permissions',
                                    to='django_ledger.entitymodel', verbose_name='Entity'),
        ),
        migrations.AddField(
            model_name='entitymanagementmodel',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entity_permissions',
                                    to=settings.AUTH_USER_MODEL, verbose_name='Manager'),
        ),
        migrations.AddField(
            model_name='customermodel',
            name='entity_model',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                    to='django_ledger.entitymodel', verbose_name='Customer Entity'),
        ),
        migrations.AddField(
            model_name='chartofaccountmodel',
            name='entity',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                    to='django_ledger.entitymodel', verbose_name='Entity'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='bill_items',
            field=models.ManyToManyField(through='django_ledger.ItemTransactionModel', to='django_ledger.itemmodel',
                                         verbose_name='Bill Items'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='cash_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='None_cash_account', to='django_ledger.accountmodel',
                                    verbose_name='Cash Account'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='ce_model',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    to='django_ledger.estimatemodel', verbose_name='Associated Customer Job/Estimate'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='ledger',
            field=models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                       to='django_ledger.ledgermodel', verbose_name='Ledger'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='prepaid_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='None_prepaid_account', to='django_ledger.accountmodel',
                                    verbose_name='Prepaid Account'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='unearned_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='None_unearned_account', to='django_ledger.accountmodel',
                                    verbose_name='Unearned Account'),
        ),
        migrations.AddField(
            model_name='billmodel',
            name='vendor',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.vendormodel',
                                    verbose_name='Vendor'),
        ),
        migrations.CreateModel(
            name='BankAccountModel',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('account_number', models.CharField(blank=True, max_length=30, null=True, validators=[
                    django.core.validators.RegexValidator(re.compile('^\\d+(?:\\d+)*\\Z'), code='invalid',
                                                          message='Only digits allowed')],
                                                    verbose_name='Account Number')),
                ('routing_number', models.CharField(blank=True, max_length=30, null=True, validators=[
                    django.core.validators.RegexValidator(re.compile('^\\d+(?:\\d+)*\\Z'), code='invalid',
                                                          message='Only digits allowed')],
                                                    verbose_name='Routing Number')),
                ('aba_number', models.CharField(blank=True, max_length=30, null=True, verbose_name='ABA Number')),
                ('swift_number', models.CharField(blank=True, max_length=30, null=True, verbose_name='SWIFT Number')),
                ('account_type',
                 models.CharField(choices=[('checking', 'Checking'), ('savings', 'Savings')], default='checking',
                                  max_length=10, verbose_name='Account Type')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, max_length=150, null=True)),
                ('active', models.BooleanField(default=False)),
                ('hidden', models.BooleanField(default=False)),
                ('cash_account',
                 models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, related_name='bank_cash_account',
                                   to='django_ledger.accountmodel', verbose_name='Cash Account')),
                ('entity_model',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel',
                                   verbose_name='Entity Model')),
            ],
            options={
                'verbose_name': 'Bank Account',
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='accountmodel',
            name='coa_model',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE,
                                    to='django_ledger.chartofaccountmodel', verbose_name='Chart of Accounts'),
        ),
        migrations.AddIndex(
            model_name='vendormodel',
            index=models.Index(fields=['created'], name='django_ledg_created_6a5d9d_idx'),
        ),
        migrations.AddIndex(
            model_name='vendormodel',
            index=models.Index(fields=['updated'], name='django_ledg_updated_736f82_idx'),
        ),
        migrations.AddIndex(
            model_name='vendormodel',
            index=models.Index(fields=['active'], name='django_ledg_active_25e3b1_idx'),
        ),
        migrations.AddIndex(
            model_name='vendormodel',
            index=models.Index(fields=['hidden'], name='django_ledg_hidden_379b0a_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='vendormodel',
            unique_together={('entity_model', 'vendor_number')},
        ),
        migrations.AddIndex(
            model_name='unitofmeasuremodel',
            index=models.Index(fields=['entity'], name='django_ledg_entity__1c8986_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='unitofmeasuremodel',
            unique_together={('entity', 'unit_abbr')},
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['tx_type'], name='django_ledg_tx_type_da7ba9_idx'),
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['account'], name='django_ledg_account_c4bb7e_idx'),
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['journal_entry'], name='django_ledg_journal_46c77f_idx'),
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['created'], name='django_ledg_created_b74538_idx'),
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['updated'], name='django_ledg_updated_494252_idx'),
        ),
        migrations.AddIndex(
            model_name='stagedtransactionmodel',
            index=models.Index(fields=['import_job'], name='django_ledg_import__8e6511_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['entity'], name='django_ledg_entity__9ab6ba_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['po_number'], name='django_ledg_po_numb_20b7d1_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['po_status'], name='django_ledg_po_stat_ab8d41_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['ce_model'], name='django_ledg_ce_mode_66c2cc_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['date_draft'], name='django_ledg_date_dr_6fd15b_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['date_in_review'], name='django_ledg_date_in_b6a1f3_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['date_approved'], name='django_ledg_date_ap_4667c6_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['date_fulfilled'], name='django_ledg_date_fu_b29b08_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['date_canceled'], name='django_ledg_date_ca_3edc4b_idx'),
        ),
        migrations.AddIndex(
            model_name='purchaseordermodel',
            index=models.Index(fields=['date_void'], name='django_ledg_date_vo_56612f_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='purchaseordermodel',
            unique_together={('entity', 'po_number')},
        ),
        migrations.AddIndex(
            model_name='ledgermodel',
            index=models.Index(fields=['entity'], name='django_ledg_entity__e21c5d_idx'),
        ),
        migrations.AddIndex(
            model_name='ledgermodel',
            index=models.Index(fields=['entity', 'posted'], name='django_ledg_entity__040422_idx'),
        ),
        migrations.AddIndex(
            model_name='ledgermodel',
            index=models.Index(fields=['entity', 'locked'], name='django_ledg_entity__cde962_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['ledger'], name='django_ledg_ledger__ecaa89_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['timestamp'], name='django_ledg_timesta_ab91e8_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['activity'], name='django_ledg_activit_bae3bb_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['entity_unit'], name='django_ledg_entity__557f42_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['locked'], name='django_ledg_locked_5cc524_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['posted'], name='django_ledg_posted_097e54_idx'),
        ),
        migrations.AddIndex(
            model_name='journalentrymodel',
            index=models.Index(fields=['je_number'], name='django_ledg_je_numb_987c8b_idx'),
        ),
        migrations.AddIndex(
            model_name='itemtransactionmodel',
            index=models.Index(fields=['bill_model', 'item_model'], name='django_ledg_bill_mo_479f1f_idx'),
        ),
        migrations.AddIndex(
            model_name='itemtransactionmodel',
            index=models.Index(fields=['invoice_model', 'item_model'], name='django_ledg_invoice_13ac3b_idx'),
        ),
        migrations.AddIndex(
            model_name='itemtransactionmodel',
            index=models.Index(fields=['po_model', 'item_model'], name='django_ledg_po_mode_2675c9_idx'),
        ),
        migrations.AddIndex(
            model_name='itemtransactionmodel',
            index=models.Index(fields=['ce_model', 'item_model'], name='django_ledg_ce_mode_ae4efd_idx'),
        ),
        migrations.AddIndex(
            model_name='itemtransactionmodel',
            index=models.Index(fields=['po_item_status'], name='django_ledg_po_item_ac9475_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['inventory_account'], name='django_ledg_invento_dbf206_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['cogs_account'], name='django_ledg_cogs_ac_82d441_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['earnings_account'], name='django_ledg_earning_229a60_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['expense_account'], name='django_ledg_expense_f65128_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['for_inventory'], name='django_ledg_for_inv_c93303_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['is_product_or_service'], name='django_ledg_is_prod_4b9941_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['is_active'], name='django_ledg_is_acti_ffc9c6_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['item_type'], name='django_ledg_item_ty_0cef58_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['sku'], name='django_ledg_sku_02b4d8_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['upc'], name='django_ledg_upc_c19d0a_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['item_id'], name='django_ledg_item_id_972760_idx'),
        ),
        migrations.AddIndex(
            model_name='itemmodel',
            index=models.Index(fields=['item_number'], name='django_ledg_item_nu_1e3c20_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='itemmodel',
            unique_together={('entity', 'item_number')},
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['invoice_status'], name='django_ledg_invoice_e349d0_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['terms'], name='django_ledg_terms_3b6577_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['cash_account'], name='django_ledg_cash_ac_00d697_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['prepaid_account'], name='django_ledg_prepaid_470f67_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['unearned_account'], name='django_ledg_unearne_bc18fd_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_due'], name='django_ledg_date_du_c72892_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_draft'], name='django_ledg_date_dr_ad2569_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_in_review'], name='django_ledg_date_in_e738b8_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_approved'], name='django_ledg_date_ap_ae90e1_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_paid'], name='django_ledg_date_pa_34b872_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_canceled'], name='django_ledg_date_ca_0f65e8_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['date_void'], name='django_ledg_date_vo_57c444_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['customer'], name='django_ledg_custome_34c881_idx'),
        ),
        migrations.AddIndex(
            model_name='invoicemodel',
            index=models.Index(fields=['invoice_number'], name='django_ledg_invoice_d0881e_idx'),
        ),
        migrations.AddIndex(
            model_name='importjobmodel',
            index=models.Index(fields=['ledger'], name='django_ledg_ledger__1e8758_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['status'], name='django_ledg_status_6aa955_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['customer'], name='django_ledg_custome_a2793e_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['terms'], name='django_ledg_terms_9507f6_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['entity'], name='django_ledg_entity__e2e967_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['date_draft'], name='django_ledg_date_dr_502e29_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['date_in_review'], name='django_ledg_date_in_b24ddd_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['date_approved'], name='django_ledg_date_ap_a754e5_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['date_canceled'], name='django_ledg_date_ca_9af3db_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['date_void'], name='django_ledg_date_vo_74c7c8_idx'),
        ),
        migrations.AddIndex(
            model_name='estimatemodel',
            index=models.Index(fields=['estimate_number'], name='django_ledg_estimat_9ff8a5_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='estimatemodel',
            unique_together={('entity', 'estimate_number')},
        ),
        migrations.AddIndex(
            model_name='entityunitmodel',
            index=models.Index(fields=['active'], name='django_ledg_active_2c2caa_idx'),
        ),
        migrations.AddIndex(
            model_name='entityunitmodel',
            index=models.Index(fields=['hidden'], name='django_ledg_hidden_a01d42_idx'),
        ),
        migrations.AddIndex(
            model_name='entityunitmodel',
            index=models.Index(fields=['entity'], name='django_ledg_entity__0bdfdc_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='entityunitmodel',
            unique_together={('entity', 'slug'), ('entity', 'document_prefix')},
        ),
        migrations.AddIndex(
            model_name='entitystatemodel',
            index=models.Index(fields=['key'], name='django_ledg_key_ac156a_idx'),
        ),
        migrations.AddIndex(
            model_name='entitystatemodel',
            index=models.Index(fields=['entity_model', 'fiscal_year', 'entity_unit', 'key'],
                               name='django_ledg_entity__eeae49_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='entitystatemodel',
            unique_together={('entity_model', 'entity_unit', 'fiscal_year', 'key')},
        ),
        migrations.AddIndex(
            model_name='entitymodel',
            index=models.Index(fields=['admin'], name='django_ledg_admin_i_09f5c9_idx'),
        ),
        migrations.AddIndex(
            model_name='entitymanagementmodel',
            index=models.Index(fields=['entity', 'user'], name='django_ledg_entity__9541e6_idx'),
        ),
        migrations.AddIndex(
            model_name='entitymanagementmodel',
            index=models.Index(fields=['user', 'entity'], name='django_ledg_user_id_b7497b_idx'),
        ),
        migrations.AddIndex(
            model_name='customermodel',
            index=models.Index(fields=['created'], name='django_ledg_created_ba7f4c_idx'),
        ),
        migrations.AddIndex(
            model_name='customermodel',
            index=models.Index(fields=['updated'], name='django_ledg_updated_f9ac90_idx'),
        ),
        migrations.AddIndex(
            model_name='customermodel',
            index=models.Index(fields=['active'], name='django_ledg_active_967a81_idx'),
        ),
        migrations.AddIndex(
            model_name='customermodel',
            index=models.Index(fields=['hidden'], name='django_ledg_hidden_dda722_idx'),
        ),
        migrations.AddIndex(
            model_name='customermodel',
            index=models.Index(fields=['customer_number'], name='django_ledg_custome_16f95a_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='customermodel',
            unique_together={('entity_model', 'customer_number')},
        ),
        migrations.AddIndex(
            model_name='chartofaccountmodel',
            index=models.Index(fields=['entity'], name='django_ledg_entity__48d6e0_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['bill_status'], name='django_ledg_bill_st_9e158c_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['terms'], name='django_ledg_terms_752251_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['cash_account'], name='django_ledg_cash_ac_82021a_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['prepaid_account'], name='django_ledg_prepaid_5230e5_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['unearned_account'], name='django_ledg_unearne_806444_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_due'], name='django_ledg_date_du_a8c481_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_draft'], name='django_ledg_date_dr_7a448e_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_in_review'], name='django_ledg_date_in_8887cb_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_approved'], name='django_ledg_date_ap_3208a1_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_paid'], name='django_ledg_date_pa_daf06c_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_canceled'], name='django_ledg_date_ca_e43055_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['date_void'], name='django_ledg_date_vo_14c747_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['vendor'], name='django_ledg_vendor__28be85_idx'),
        ),
        migrations.AddIndex(
            model_name='billmodel',
            index=models.Index(fields=['bill_number'], name='django_ledg_bill_nu_6b97b4_idx'),
        ),
        migrations.AddIndex(
            model_name='bankaccountmodel',
            index=models.Index(fields=['account_type'], name='django_ledg_account_009a4a_idx'),
        ),
        migrations.AddIndex(
            model_name='bankaccountmodel',
            index=models.Index(fields=['cash_account'], name='django_ledg_cash_ac_59a8af_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='bankaccountmodel',
            unique_together={('entity_model', 'account_number'),
                             ('entity_model', 'cash_account', 'account_number', 'routing_number')},
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['role'], name='django_ledg_role_812d08_idx'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['balance_type'], name='django_ledg_balance_daddac_idx'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['active'], name='django_ledg_active_f8adc2_idx'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['coa_model'], name='django_ledg_coa_mod_e19964_idx'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['role', 'balance_type', 'active'], name='django_ledg_role_1bff96_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='accountmodel',
            unique_together={('coa_model', 'code')},
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

This is the base Chart of Accounts that has all the possible accounts that are useful for the preparation of the
Financial Statements. A user may choose to use the default CoA at the creation of each EntityModel but it is not
required. The default CoA is intended to provide a QuickStart solution for most use cases.

The Chart of Accounts is broadly bifurcated into 5 different Sections:
    1. Assets:
    2. Liabilities
    3. Shareholder's Equity
    4. Income
    5. COGS
    6. Expenses


The Django Ledger Default Chart of Accounts must include the following fields:
    * Code - String
    * Role - A choice from any of the possible account roles (see django_ledger.roles module).
    * Balance Type - A CREDIT or DEBIT balance account setting.
    * Name - A human readable name.
    * Parent - The parent account of the AccountModel instance.

If the DEFAULT_CHART_OF_ACCOUNTS setting is present, the default CoA will be replace by such setting.

Default Chart of Accounts Table
===============================

======  ==========================  ==============  ===================================================  ========  ================
  code  role                        balance_type    name                                                   parent  root_group
======  ==========================  ==============  ===================================================  ========  ================
  1910  asset_adjustment            debit           Securities Unrealized Gains/Losses                             root_assets
  1920  asset_adjustment            debit           PPE Unrealized Gains/Losses                                    root_assets
  1010  asset_ca_cash               debit           Cash                                                           root_assets
  1050  asset_ca_mkt_sec            debit           Short Term Investments                                         root_assets
  1200  asset_ca_inv                debit           Inventory                                                      root_assets
  1300  asset_ca_prepaid            debit           Prepaid Expenses                                               root_assets
  1100  asset_ca_recv               debit           Accounts Receivable                                            root_assets
  1110  asset_ca_uncoll             credit          Uncollectibles                                                 root_assets
  1810  asset_ia                    debit           Goodwill                                                       root_assets
  1820  asset_ia                    debit           Intellectual Property                                          root_assets
  1830  asset_ia_accum_amort        credit          Less: Intangible Assets Accumulated Amortization               root_assets
  1520  asset_lti_land              debit           Land                                                           root_assets
  1510  asset_lti_notes             debit           Notes Receivable                                               root_assets
  1530  asset_lti_sec               debit           Securities                                                     root_assets
  1610  asset_ppe_build             debit           Buildings                                                      root_assets
  1611  asset_ppe_build_accum_depr  credit          Less: Buildings Accumulated Depreciation                       root_assets
  1630  asset_ppe_equip             debit           Equipment                                                      root_assets
  1631  asset_ppe_equip_accum_depr  credit          Less: Equipment Accumulated Depreciation                       root_assets
  1620  asset_ppe_plant             debit           Plant                                                          root_assets
  1640  asset_ppe_plant             debit           Vehicles                                                       root_assets
  1650  asset_ppe_plant             debit           Furniture & Fixtures                                           root_assets
  1621  asset_ppe_plant_depr        credit          Less: Plant Accumulated Depreciation                           root_assets
  1641  asset_ppe_plant_depr        credit          Less: Vehicles Accumulated Depreciation                        root_assets
  1651  asset_ppe_plant_depr        credit          Less: Furniture & Fixtures Accumulated Depreciation            root_assets
  3910  eq_adjustment               credit          Available for Sale                                             root_capital
  3920  eq_adjustment               credit          PPE Unrealized Gains/Losses                                    root_capital
  3010  eq_capital                  credit          Capital Account 1                                              root_capital
  3020  eq_capital                  credit          Capital Account 2                                              root_capital
  3030  eq_capital                  credit          Capital Account 3                                              root_capital
  3930  eq_dividends                debit           Dividends & Distributions                                      root_capital
  3110  eq_stock_common             credit          Common Stock                                                   root_capital
  3120  eq_stock_preferred          credit          Preferred Stock                                                root_capital
  5010  cogs_regular                debit           Cost of Goods Sold                                             root_cogs
  6075  ex_amortization             debit           Amortization Expense                                           root_expenses
  6070  ex_depreciation             debit           Depreciation Expense                                           root_expenses
  6130  ex_interest                 debit           Interest Expense                                               root_expenses
  6500  ex_other                    debit           Misc. Expense                                                  root_expenses
  6010  ex_regular                  debit           Advertising                                                    root_expenses
  6020  ex_regular                  debit           Amortization                                                   root_expenses
  6030  ex_regular                  debit           Auto Expense                                                   root_expenses
  6040  ex_regular                  debit           Bad Debt                                                       root_expenses
  6050  ex_regular                  debit           Bank Charges                                                   root_expenses
  6060  ex_regular                  debit           Commission Expense                                             root_expenses
  6080  ex_regular                  debit           Employee Benefits                                              root_expenses
  6081  ex_regular                  debit           Employee Wages                                                 root_expenses
  6090  ex_regular                  debit           Freight                                                        root_expenses
  6110  ex_regular                  debit           Gifts                                                          root_expenses
  6120  ex_regular                  debit           Insurance                                                      root_expenses
  6140  ex_regular                  debit           Professional Fees                                              root_expenses
  6150  ex_regular                  debit           License Expense                                                root_expenses
  6170  ex_regular                  debit           Maintenance Expense                                            root_expenses
  6180  ex_regular                  debit           Meals & Entertainment                                          root_expenses
  6190  ex_regular                  debit           Office Expense                                                 root_expenses
  6220  ex_regular                  debit           Printing                                                       root_expenses
  6230  ex_regular                  debit           Postage                                                        root_expenses
  6240  ex_regular                  debit           Rent                                                           root_expenses
  6250  ex_regular                  debit           Maintenance & Repairs                                          root_expenses
  6251  ex_regular                  debit           Maintenance                                                    root_expenses
  6252  ex_regular                  debit           Repairs                                                        root_expenses
  6253  ex_regular                  debit           HOA                                                            root_expenses
  6254  ex_regular                  debit           Snow Removal                                                   root_expenses
  6255  ex_regular                  debit           Lawn Care                                                      root_expenses
  6260  ex_regular                  debit           Salaries                                                       root_expenses
  6270  ex_regular                  debit           Supplies                                                       root_expenses
  6290  ex_regular                  debit           Utilities                                                      root_expenses
  6292  ex_regular                  debit           Sewer                                                          root_expenses
  6293  ex_regular                  debit           Gas                                                            root_expenses
  6294  ex_regular                  debit           Garbage                                                        root_expenses
  6295  ex_regular                  debit           Electricity                                                    root_expenses
  6300  ex_regular                  debit           Property Management                                            root_expenses
  6400  ex_regular                  debit           Vacancy                                                        root_expenses
  6210  ex_taxes                    debit           Payroll Taxes                                                  root_expenses
  6280  ex_taxes                    debit           Taxes                                                          root_expenses
  4040  in_gain_loss                credit          Capital Gain/Loss Income                                       root_income
  4030  in_interest                 credit          Interest Income                                                root_income
  4010  in_operational              credit          Sales Income                                                   root_income
  4050  in_other                    credit          Other Income                                                   root_income
  4020  in_passive                  credit          Investing Income                                               root_income
  2010  lia_cl_acc_payable          credit          Accounts Payable                                               root_liabilities
  2060  lia_cl_def_rev              credit          Deferred Revenues                                              root_liabilities
  2030  lia_cl_int_payable          credit          Interest Payable                                               root_liabilities
  2050  lia_cl_ltd_mat              credit          Current Maturities LT Debt                                     root_liabilities
  2070  lia_cl_other                credit          Other Payables                                                 root_liabilities
  2040  lia_cl_st_notes_payable     credit          Short-Term Notes Payable                                       root_liabilities
  2020  lia_cl_wages_payable        credit          Wages Payable                                                  root_liabilities
  2120  lia_ltl_bonds               credit          Bonds Payable                                                  root_liabilities
  2130  lia_ltl_mortgage            credit          Mortgage Payable                                               root_liabilities
  2110  lia_ltl_notes               credit          Long Term Notes Payable                                        root_liabilities
======  ==========================  ==============  ===================================================  ========  ================
"""

from itertools import groupby
from typing import Optional, Dict, List

from django_ledger.exceptions import DjangoLedgerConfigurationError
from django_ledger.io import roles, ROOT_ASSETS, ROOT_INCOME, ROOT_EXPENSES, ROOT_LIABILITIES, ROOT_CAPITAL, ROOT_COGS
from django_ledger.settings import DJANGO_LEDGER_DEFAULT_COA

# todo: include a function to use a user-defined CHART_OF_ACCOUNTS option.

DEFAULT_CHART_OF_ACCOUNTS = [

    # ---------# ASSETS START #---------#
    # CURRENT ASSETS  ------
    {'code': '1010', 'role': roles.ASSET_CA_CASH, 'balance_type': 'debit', 'name': 'Cash', 'parent': None},
    {'code': '1050', 'role': roles.ASSET_CA_MKT_SECURITIES, 'balance_type': 'debit', 'name': 'Short Term Investments',
     'parent': None},
    {'code': '1100', 'role': roles.ASSET_CA_RECEIVABLES, 'balance_type': 'debit', 'name': 'Accounts Receivable',
     'parent': None},
    {'code': '1110', 'role': roles.ASSET_CA_UNCOLLECTIBLES, 'balance_type': 'credit', 'name': 'Uncollectibles',
     'parent': None},
    {'code': '1200', 'role': roles.ASSET_CA_INVENTORY, 'balance_type': 'debit', 'name': 'Inventory', 'parent': None},
    {'code': '1300', 'role': roles.ASSET_CA_PREPAID, 'balance_type': 'debit', 'name': 'Prepaid Expenses',
     'parent': None},

    # LONG TERM INVESTMENTS ------
    {'code': '1510', 'role': roles.ASSET_LTI_NOTES_RECEIVABLE, 'balance_type': 'debit', 'name': 'Notes Receivable',
     'parent': None},
    {'code': '1520', 'role': roles.ASSET_LTI_LAND, 'balance_type': 'debit', 'name': 'Land', 'parent': None},
    {'code': '1530', 'role': roles.ASSET_LTI_SECURITIES, 'balance_type': 'debit', 'name': 'Securities', 'parent': None},

    # PPE ------
    {'code': '1610', 'role': roles.ASSET_PPE_BUILDINGS, 'balance_type': 'debit', 'name': 'Buildings', 'parent': None},
    {'code': '1611', 'role': roles.ASSET_PPE_BUILDINGS_ACCUM_DEPRECIATION, 'balance_type': 'credit',
     'name': 'Less: Buildings Accumulated Depreciation', 'parent': None},
    {'code': '1620', 'role': roles.ASSET_PPE_PLANT, 'balance_type': 'debit', 'name': 'Plant', 'parent': None},
    {'code': '1621', 'role': roles.ASSET_PPE_PLANT_ACCUM_DEPRECIATION, 'balance_type': 'credit',
     'name': 'Less: Plant Accumulated Depreciation', 'parent': None},
    {'code': '1630', 'role': roles.ASSET_PPE_EQUIPMENT, 'balance_type': 'debit', 'name': 'Equipment', 'parent': None},
    {'code': '1631', 'role': roles.ASSET_PPE_EQUIPMENT_ACCUM_DEPRECIATION, 'balance_type': 'credit',
     'name': 'Less: Equipment Accumulated Depreciation', 'parent': None},
    {'code': '1640', 'role': roles.ASSET_PPE_PLANT, 'balance_type': 'debit', 'name': 'Vehicles', 'parent': None},
    {'code': '1641', 'role': roles.ASSET_PPE_PLANT_ACCUM_DEPRECIATION, 'balance_type': 'credit',
     'name': 'Less: Vehicles Accumulated Depreciation', 'parent': None},
    {'code': '1650', 'role': roles.ASSET_PPE_PLANT, 'balance_type': 'debit', 'name': 'Furniture & Fixtures',
     'parent': None},
    {'code': '1651', 'role': roles.ASSET_PPE_PLANT_ACCUM_DEPRECIATION, 'balance_type': 'credit',
     'name': 'Less: Furniture & Fixtures Accumulated Depreciation', 'parent': None},

    # INTANGIBLE ASSETS ------
    {'code': '1810', 'role': roles.ASSET_INTANGIBLE_ASSETS, 'balance_type': 'debit', 'name': 'Goodwill',
     'parent': None},
    {'code': '1820', 'role': roles.ASSET_INTANGIBLE_ASSETS, 'balance_type': 'debit', 'name': 'Intellectual Property',
     'parent': None},
    {'code': '1830', 'role': roles.ASSET_INTANGIBLE_ASSETS_ACCUM_AMORTIZATION, 'balance_type': 'credit',
     'name': 'Less: Intangible Assets Accumulated Amortization', 'parent': '1820'},

    # ADJUSTMENTS ------
    {'code': '1910', 'role': roles.ASSET_ADJUSTMENTS, 'balance_type': 'debit',
     'name': 'Securities Unrealized Gains/Losses', 'parent': None},
    {'code': '1920', 'role': roles.ASSET_ADJUSTMENTS, 'balance_type': 'debit', 'name': 'PPE Unrealized Gains/Losses',
     'parent': None},

    # ---------# ASSETS END #---------#

    # ---------# LIABILITIES START #---------#
    # CURRENT LIABILITIES ------
    {'code': '2010', 'role': roles.LIABILITY_CL_ACC_PAYABLE, 'balance_type': 'credit', 'name': 'Accounts Payable',
     'parent': None},
    {'code': '2020', 'role': roles.LIABILITY_CL_WAGES_PAYABLE, 'balance_type': 'credit', 'name': 'Wages Payable',
     'parent': None},
    {'code': '2030', 'role': roles.LIABILITY_CL_INTEREST_PAYABLE, 'balance_type': 'credit', 'name': 'Interest Payable',
     'parent': None},
    {'code': '2040', 'role': roles.LIABILITY_CL_ST_NOTES_PAYABLE, 'balance_type': 'credit',
     'name': 'Short-Term Notes Payable', 'parent': None},
    {'code': '2050', 'role': roles.LIABILITY_CL_LTD_MATURITIES, 'balance_type': 'credit',
     'name': 'Current Maturities LT Debt', 'parent': None},
    {'code': '2060', 'role': roles.LIABILITY_CL_DEFERRED_REVENUE, 'balance_type': 'credit', 'name': 'Deferred Revenues',
     'parent': None},
    {'code': '2070', 'role': roles.LIABILITY_CL_OTHER, 'balance_type': 'credit', 'name': 'Other Payables',
     'parent': None},

    # LIABILITIES ACCOUNTS ------
    {'code': '2110', 'role': roles.LIABILITY_LTL_NOTES_PAYABLE, 'balance_type': 'credit',
     'name': 'Long Term Notes Payable', 'parent': None},
    {'code': '2120', 'role': roles.LIABILITY_LTL_BONDS_PAYABLE, 'balance_type': 'credit', 'name': 'Bonds Payable',
     'parent': None},
    {'code': '2130', 'role': roles.LIABILITY_LTL_MORTGAGE_PAYABLE, 'balance_type': 'credit', 'name': 'Mortgage Payable',
     'parent': None},

    # ---------# LIABILITIES END #---------#

    # ---------# SHEREHOLDERS EQUITY START #---------#
    # CAPITAL ACCOUNTS ------
    {'code': '3010', 'role': roles.EQUITY_CAPITAL, 'balance_type': 'credit', 'name': 'Capital Account 1',
     'parent': None},
    {'code': '3020', 'role': roles.EQUITY_CAPITAL, 'balance_type': 'credit', 'name': 'Capital Account 2',
     'parent': None},
    {'code': '3030', 'role': roles.EQUITY_CAPITAL, 'balance_type': 'credit', 'name': 'Capital Account 3',
     'parent': None},

    {'code': '3110', 'role': roles.EQUITY_COMMON_STOCK, 'balance_type': 'credit', 'name': 'Common Stock',
     'parent': None},
    {'code': '3120', 'role': roles.EQUITY_PREFERRED_STOCK, 'balance_type': 'credit', 'name': 'Preferred Stock',
     'parent': None},

    {'code': '3910', 'role': roles.EQUITY_ADJUSTMENT, 'balance_type': 'credit', 'name': 'Available for Sale',
     'parent': None},
    {'code': '3920', 'role': roles.EQUITY_ADJUSTMENT, 'balance_type': 'credit', 'name': 'PPE Unrealized Gains/Losses',
     'parent': None},

    {'code': '3930', 'role': roles.EQUITY_DIVIDENDS, 'balance_type': 'debit', 'name': 'Dividends & Distributions',
     'parent': None},

    # REVENUE ACCOUNTS ------
    {'code': '4010', 'role': roles.INCOME_OPERATIONAL, 'balance_type': 'credit', 'name': 'Sales Income',
     'parent': None},
    {'code': '4020', 'role': roles.INCOME_PASSIVE, 'balance_type': 'credit', 'name': 'Investing Income',
     'parent': None},
    {'code': '4030', 'role': roles.INCOME_INTEREST, 'balance_type': 'credit', 'name': 'Interest Income',
     'parent': None},
    {'code': '4040', 'role': roles.INCOME_CAPITAL_GAIN_LOSS, 'balance_type': 'credit',
     'name': 'Capital Gain/Loss Income', 'parent': None},
    {'code': '4050', 'role': roles.INCOME_OTHER, 'balance_type': 'credit', 'name': 'Other Income', 'parent': None},

    # COGS ACCOUNTS ------
    {'code': '5010', 'role': roles.COGS, 'balance_type': 'debit', 'name': 'Cost of Goods Sold', 'parent': None},

    # EXPENSE ACCOUNTS ------
    {'code': '6010', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Advertising', 'parent': None},
    {'code': '6020', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Amortization',
     'parent': None},
    {'code': '6030', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Auto Expense',
     'parent': None},
    {'code': '6040', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Bad Debt', 'parent': None},
    {'code': '6050', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Bank Charges',
     'parent': None},
    {'code': '6060', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Commission Expense',
     'parent': None},
    {'code': '6080', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Employee Benefits',
     'parent': None},
    {'code': '6081', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Employee Wages',
     'parent': None},
    {'code': '6090', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Freight', 'parent': None},
    {'code': '6110', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Gifts', 'parent': None},
    {'code': '6120', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Insurance', 'parent': None},
    {'code': '6140', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Professional Fees',
     'parent': None},
    {'code': '6150', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'License Expense',
     'parent': None},
    {'code': '6170', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Maintenance Expense',
     'parent': None},
    {'code': '6180', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Meals & Entertainment',
     'parent': None},
    {'code': '6190', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Office Expense',
     'parent': None},
    {'code': '6220', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Printing', 'parent': None},
    {'code': '6230', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Postage', 'parent': None},
    {'code': '6240', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Rent', 'parent': None},
    {'code': '6250', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Maintenance & Repairs',
     'parent': None},
    {'code': '6251', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Maintenance', 'parent': None},
    {'code': '6252', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Repairs', 'parent': None},
    {'code': '6253', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'HOA', 'parent': None},
    {'code': '6254', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Snow Removal',
     'parent': None},
    {'code': '6255', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Lawn Care', 'parent': None},
    {'code': '6260', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Salaries', 'parent': None},
    {'code': '6270', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Supplies', 'parent': None},
    {'code': '6290', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Utilities', 'parent': None},
    {'code': '6292', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Sewer', 'parent': None},
    {'code': '6293', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Gas', 'parent': None},
    {'code': '6294', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Garbage', 'parent': None},
    {'code': '6295', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Electricity', 'parent': None},
    {'code': '6300', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Property Management',
     'parent': None},
    {'code': '6400', 'role': roles.EXPENSE_OPERATIONAL, 'balance_type': 'debit', 'name': 'Vacancy', 'parent': None},

    {'code': '6070', 'role': roles.EXPENSE_DEPRECIATION, 'balance_type': 'debit', 'name': 'Depreciation Expense',
     'parent': None},
    {'code': '6075', 'role': roles.EXPENSE_AMORTIZATION, 'balance_type': 'debit', 'name': 'Amortization Expense',
     'parent': None},
    {'code': '6130', 'role': roles.EXPENSE_INTEREST_ST, 'balance_type': 'debit',
     'name': 'Interest Expense on Short Term Debt', 'parent': None},
    {'code': '6131', 'role': roles.EXPENSE_INTEREST_LT, 'balance_type': 'debit',
     'name': 'Interest Expense on Long Term Debt', 'parent': None},
    {'code': '6210', 'role': roles.EXPENSE_TAXES, 'balance_type': 'debit', 'name': 'Payroll Taxes', 'parent': None},
    {'code': '6280', 'role': roles.EXPENSE_TAXES, 'balance_type': 'debit', 'name': 'Taxes', 'parent': None},
    {'code': '6500', 'role': roles.EXPENSE_OTHER, 'balance_type': 'debit', 'name': 'Misc. Expense', 'parent': None}

]


def get_default_coa() -> List[Dict]:
    if DJANGO_LEDGER_DEFAULT_COA is not None and isinstance(DJANGO_LEDGER_DEFAULT_COA, list):
        return DJANGO_LEDGER_DEFAULT_COA
    return DEFAULT_CHART_OF_ACCOUNTS


if DJANGO_LEDGER_DEFAULT_COA:
    DJANGO_LEDGER_DEFAULT_COA = get_default_coa()

PREFIX_MAP = {
    'in': ROOT_INCOME,
    'ex': ROOT_EXPENSES,
    'lia': ROOT_LIABILITIES,
    'eq': ROOT_CAPITAL,
    'asset': ROOT_ASSETS,
    'cogs': ROOT_COGS
}

for i in DEFAULT_CHART_OF_ACCOUNTS:
    i['root_group'] = PREFIX_MAP[i['role'].split('_')[0]]

DEFAULT_CHART_OF_ACCOUNTS.sort(key=lambda x: (x['root_group'], x['role'], x['code']))
CHART_OF_ACCOUNTS_ROOT_MAP = {
    k: list(v) for k, v in groupby(DEFAULT_CHART_OF_ACCOUNTS, key=lambda x: x['root_group'])
}


def verify_unique_code():
    """
    A function that verifies that there are no duplicate code in the Default CoA during the development and launch.
    """
    code_list = [i['code'] for i in DEFAULT_CHART_OF_ACCOUNTS]
    code_set = set(code_list)
    if not len(code_list) == len(code_set):
        raise DjangoLedgerConfigurationError('Default CoA is not unique.')


def get_default_coa_rst(default_coa: Optional[Dict] = None) -> str:
    """
    Converts the provided Chart of Account into restructuredText format.
    Parameters
    ----------
    default_coa:
        A dictionary of chart of accounts. Must follow the same keys as CHART_OF_ACCOUNTS.

    Returns
    -------
    str:
        The table in RestructuredText format.
    """
    try:
        from tabulate import tabulate
    except ModuleNotFoundError as e:
        raise DjangoLedgerConfigurationError(e.msg)
    if default_coa:
        return tabulate(default_coa, headers='keys', tablefmt='rst')
    return tabulate(get_default_coa(), headers='keys', tablefmt='rst')


verify_unique_code()



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""


from django.contrib import messages
from django.http import HttpResponseBadRequest, HttpResponseNotFound, HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, DetailView

from django_ledger.models import EntityModel
from django_ledger.models.items import ItemTransactionModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class InventoryListView(DjangoLedgerSecurityMixIn, ListView):
    template_name = 'django_ledger/inventory/inventory_list.html'
    context_object_name = 'inventory_list'
    http_method_names = ['get']

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super(InventoryListView, self).get_context_data(**kwargs)
        qs = self.get_queryset()

        # evaluates the queryset...
        context['qs_count'] = qs.count()

        # ordered inventory...
        ordered_qs = qs.is_ordered()
        context['inventory_ordered'] = ordered_qs

        # in transit inventory...
        in_transit_qs = qs.in_transit()
        context['inventory_in_transit'] = in_transit_qs

        # on hand inventory...
        received_qs = qs.is_received()
        context['inventory_received'] = received_qs

        context['page_title'] = _('Inventory')
        context['header_title'] = _('Inventory Status')
        context['header_subtitle'] = _('Ordered/In Transit/On Hand')
        context['header_subtitle_icon'] = 'ic:round-inventory'
        return context

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = ItemTransactionModel.objects.inventory_pipeline_aggregate(
                entity_model=self.AUTHORIZED_ENTITY_MODEL
            )
        return super().get_queryset()


class InventoryRecountView(DjangoLedgerSecurityMixIn, DetailView):
    template_name = 'django_ledger/inventory/inventory_recount.html'
    http_method_names = ['get']
    slug_url_kwarg = 'entity_slug'

    def get_queryset(self):
        if not self.queryset:
            self.queryset = EntityModel.objects.for_user(
                user_model=self.request.user
            )
        return super().get_queryset()

    def counted_inventory(self):
        return ItemTransactionModel.objects.inventory_count(entity_model=self.AUTHORIZED_ENTITY_MODEL)

    def recorded_inventory(self, queryset=None, as_values=True):
        entity_model: EntityModel = self.AUTHORIZED_ENTITY_MODEL
        recorded_qs = entity_model.recorded_inventory(item_qs=queryset)
        return recorded_qs

    def get_context_data(self, adjustment=None, counted_qs=None, recorded_qs=None, **kwargs):
        self.object = self.get_object()
        context = super(InventoryRecountView, self).get_context_data(**kwargs)
        context['page_title'] = _('Inventory Recount')
        context['header_title'] = _('Inventory Recount')

        recorded_qs = self.recorded_inventory() if not recorded_qs else recorded_qs
        counted_qs = self.counted_inventory() if not counted_qs else counted_qs
        adjustment = EntityModel.inventory_adjustment(counted_qs, recorded_qs) if not adjustment else adjustment

        context['count_inventory_received'] = counted_qs
        context['current_inventory_levels'] = recorded_qs
        context['inventory_adjustment'] = [(k, v) for k, v in adjustment.items() if any(v.values())]

        return context

    def get(self, request, *args, **kwargs):

        confirm = self.request.GET.get('confirm')

        if confirm:
            try:
                confirm = int(confirm)
            except TypeError:
                return HttpResponseBadRequest('Not Found. Invalid conform code...')
            finally:
                if confirm not in [0, 1]:
                    return HttpResponseNotFound('Not Found. Invalid conform code...')

            self.update_inventory()
            messages.add_message(
                request,
                level=messages.INFO,
                message=f'Successfully updated recorded inventory.',
                extra_tags='is-success'
            )
            return HttpResponseRedirect(
                redirect_to=reverse('django_ledger:inventory-recount',
                                    kwargs={
                                        'entity_slug': self.kwargs['entity_slug']
                                    })
            )
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def update_inventory(self):
        entity_model: EntityModel = self.get_object()
        adj, counted_qs, recorded_qs = entity_model.update_inventory(commit=True)
        return adj, counted_qs, recorded_qs



from django_ledger.models.mixins import *
from django_ledger.models.bank_account import *
from django_ledger.models.chart_of_accounts import *
from django_ledger.models.bill import *
from django_ledger.models.invoice import *
from django_ledger.models.items import *
from django_ledger.models.accounts import *
from django_ledger.models.ledger import *
from django_ledger.models.transactions import *
from django_ledger.models.journal_entry import *
from django_ledger.models.customer import *
from django_ledger.models.estimate import *
from django_ledger.models.vendor import *
from django_ledger.models.unit import *
from django_ledger.models.purchase_order import *
from django_ledger.models.closing_entry import *
from django_ledger.models.entity import *
from django_ledger.models.data_import import *


"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>

This module contains classes and functions used to document, dispatch and commit new transaction into the database.
"""
import enum
from collections import defaultdict
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from itertools import chain
from typing import Union, Dict, Callable, Optional, List, Set
from uuid import UUID

from django.core.exceptions import ValidationError
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import get_localtime
from django_ledger.models.accounts import AccountModel, AccountModelQuerySet, CREDIT, DEBIT
from django_ledger.models.chart_of_accounts import ChartOfAccountModel
from django_ledger.models.entity import EntityModel
from django_ledger.models.ledger import LedgerModel, LedgerModelQuerySet


@dataclass
class TransactionInstructionItem:
    """
    A class to represent a transaction instruction used during the development of transaction blueprints.

    Attributes
    ----------
    account_code: str
        The account code of the AccountModel as a String.
    amount: Decimal
        The transaction amount as a Decimal value. Will be rounded to the nearest decimal place.
    tx_type: str
        A choice of 'debit' or 'credit' transaction.
    description: str
        Description of the transaction.
    account_model: AccountModel
        The resolved account model for the transaction. Not to be modified. Defaults to None.
    """
    account_code: str
    amount: Union[Decimal, float]
    tx_type: str
    description: Optional[str]
    account_model: Optional[AccountModel] = None

    def to_dict(self) -> Dict:
        return {
            'account': self.account_model,
            'amount': self.amount,
            'tx_type': self.tx_type,
            'description': self.description
        }


class IOCursorValidationError(ValidationError):
    pass


class IOCursorMode(enum.Enum):
    STRICT = 'strict'
    PERMISSIVE = 'permissive'


class IOCursor:
    """
    Represents a Django Ledger cursor capable of dispatching transactions to the database.
    The Cursor class is responsible for coordinating the creation of new ledgers, journal entries and transactions
    It is a low level interface to the IOBlueprint and IOLibrary classes.

    Parameters
    ----------
    io_library: IOLibrary
        The IOLibrary class that contains all the necessary instructions to dispatch the transactions.
    entity_model: EntityModel
        The EntityModel instance that will be used for the new transactions.
    user_model: UserModel
        The UserModel instance that will be used for the new transactions. Used for read permissions of QuerySets.
    coa_model: ChartOfAccountModel or UUID or str.
        The ChartOfAccountModel instance that contains the accounts to be used for transactions.
        Instance, UUID or slug can be sued to retrieve the model.
    """

    def __init__(self,
                 io_library,
                 entity_model: EntityModel,
                 user_model,
                 mode: IOCursorMode = IOCursorMode.PERMISSIVE,
                 coa_model: Optional[Union[ChartOfAccountModel, UUID, str]] = None):
        self.IO_LIBRARY = io_library
        self.MODE = mode
        self.ENTITY_MODEL = entity_model
        self.USER_MODEL = user_model
        self.COA_MODEL = coa_model
        self.blueprints = defaultdict(list)
        self.ledger_model_qs: Optional[LedgerModelQuerySet] = None
        self.account_model_qs: Optional[AccountModelQuerySet] = None
        self.ledger_map = dict()
        self.commit_plan = dict()
        self.instructions = None
        self.__COMMITTED: bool = False

    def get_ledger_model_qs(self) -> LedgerModelQuerySet:
        """
        Determines the ledger model queryset associated with the entity model and user model provided.

        Returns
        -------
        LedgerModelQuerySet
        """
        return LedgerModel.objects.for_entity(
            entity_model=self.ENTITY_MODEL
        )

    def get_account_model_qs(self) -> AccountModelQuerySet:
        """
        Determines the AccountModelQuerySet associated with the Chart of Accounts specified.

        Returns
        -------
        AccountModelQuerySet
        """
        return self.ENTITY_MODEL.get_coa_accounts(
            coa_model=self.COA_MODEL
        ).can_transact()

    def resolve_account_model_qs(self, codes: Set[str]) -> AccountModelQuerySet:
        """
        Resolves the final AccountModelQuerySet associated with the given account codes used by the blueprint.

        Parameters
        ----------
        codes: List[str]
            List of codes used during the execution of the blueprint.


        Returns
        -------
        AccountModelQuerySet
            The resolved AccountModelQuerySet associated with the given codes.
        """
        if self.account_model_qs is None:
            qs = self.get_account_model_qs()
            qs = qs.filter(code__in=codes)
            self.account_model_qs = qs
        return self.account_model_qs

    def resolve_ledger_model_qs(self) -> LedgerModelQuerySet:
        """
        Resolves the final LedgerModelQuerySet associated with the provided ledger model identifiers used by the
        blueprints.

        Returns
        -------
        LedgerModelQuerySet
            The resolved LedgerModelQuerySet associated with the given ledger model identifiers.
        """
        if self.ledger_model_qs is None:
            qs = self.get_ledger_model_qs()
            by_uuid = [k for k in self.blueprints.keys() if isinstance(k, UUID)]
            by_xid = [k for k in self.blueprints.keys() if isinstance(k, str)]
            self.ledger_model_qs = qs.filter(
                Q(uuid__in=by_uuid) | Q(ledger_xid__in=by_xid)
            )
        return self.ledger_model_qs

    def is_permissive(self) -> bool:
        return self.MODE == IOCursorMode.PERMISSIVE

    def is_strict(self) -> bool:
        return self.MODE == IOCursorMode.STRICT

    def dispatch(self,
                 name,
                 ledger_model: Optional[Union[str, LedgerModel, UUID]] = None,
                 **kwargs):
        """
        Stages the instructions to be processed by the IOCursor class. This method does not commit the transactions
        into the database.

        Parameters
        ----------
        name: str
            The registered blueprint name to be staged.
        ledger_model: Optional[Union[str, LedgerModel, UUID]]
            Optional ledger model identifier to house the transactions associated with the blueprint. If none is
            provided, a new ledger model will be created.
        kwargs
            The keyword arguments to be passed to the blueprint function.
        """

        if ledger_model is not None:
            if not isinstance(ledger_model, (str, UUID, LedgerModel)):
                raise IOCursorValidationError(
                    message=_('Ledger Model must be a string or UUID or LedgerModel')
                )

            if isinstance(ledger_model, LedgerModel):
                self.ENTITY_MODEL.validate_ledger_model_for_entity(ledger_model)

        blueprint_func = self.IO_LIBRARY.get_blueprint(name)
        blueprint_txs = blueprint_func(**kwargs)
        self.blueprints[ledger_model].append(blueprint_txs)

    def compile_instructions(self) -> Dict:
        """
        Compiles the blueprint instructions into Journal Entries and Transactions to be committed to the ledger.

        Returns
        -------
        Dict
            A dictionary containing the compiled instructions.
        """
        if self.instructions is None:
            instructions = {
                ledger_model: list(chain.from_iterable(
                    io_blueprint.registry for io_blueprint in instructions
                )) for ledger_model, instructions in self.commit_plan.items()
            }

            for ledger_model, txs in instructions.items():
                total_credits = sum(t.amount for t in txs if t.tx_type == CREDIT)
                total_debits = sum(t.amount for t in txs if t.tx_type == DEBIT)

                if total_credits != total_debits:
                    raise IOCursorValidationError(
                        message=_('Total transactions Credits and Debits must be equal. '
                                  'Got CREDITs: {} and DEBITs: {}.'.format(total_credits, total_debits))
                    )

            self.instructions = instructions
        return self.instructions

    def is_committed(self) -> bool:
        """
        Determines if the IOCursor instance has committed the transactions into the database.
        A cursor can only commit transactions once.

        Returns
        -------
        bool
            True if committed, False otherwise.
        """
        return self.__COMMITTED

    def commit(self,
               je_timestamp: Optional[Union[datetime, date, str]] = None,
               je_description: Optional[str] = None,
               post_new_ledgers: bool = False,
               post_journal_entries: bool = False,
               **kwargs):

        """
        Commits the compiled blueprint transactions into the database. This action is irreversible and if journal
        entries are posted, the books will be immediately impacted by the new transactions. It is encouraged NOT to
        post anything until the transaction is reviews for accuracy.

        Parameters
        ----------
        je_timestamp: Optional[Union[datetime, date, str]]
            The date or timestamp used for the committed journal entries. If none, localtime will be used.
        je_description: Optional[str]
            The description of the journal entries. If none, no description will be used.
        post_new_ledgers: bool
            If a new ledger is created, the ledger model will be posted to the database.
        post_journal_entries: bool
            If new journal entries are created, the journal entry models will be posted to the database.
        kwargs
            Additional keyword arguments passed to the IO commit_txs function.
        """
        if self.is_committed():
            raise IOCursorValidationError(
                message=_('Transactions already committed')
            )
        qs = self.resolve_ledger_model_qs()
        self.ledger_map = {l.ledger_xid: l for l in qs if l.ledger_xid} | {l.uuid: l for l in qs}

        # checks for any locked ledgers...
        for k, ledger_model in self.ledger_map.items():
            if ledger_model.is_locked():
                raise IOCursorValidationError(
                    message=_(f'Cannot transact on a locked ledger: {ledger_model}')
                )

        for k, txs in self.blueprints.items():
            if k is None:

                if self.is_permissive():
                    # no specified xid, ledger or UUID... create one...
                    self.commit_plan[
                        self.ENTITY_MODEL.create_ledger(
                            name='Blueprint Commitment',
                            commit=False,
                            posted=post_new_ledgers
                        )
                    ] = txs
                else:
                    raise IOCursorValidationError(
                        message=_('Cannot commit transactions to a non-existing ledger')
                    )

            elif isinstance(k, str):
                try:
                    # ledger with xid already exists...
                    self.commit_plan[self.ledger_map[k]] = txs
                except KeyError:
                    if self.is_permissive():
                        # create ledger with xid provided...
                        self.commit_plan[
                            self.ENTITY_MODEL.create_ledger(
                                name=f'Blueprint Commitment {k}',
                                ledger_xid=k,
                                commit=False,
                                posted=post_new_ledgers
                            )
                        ] = txs
                    else:
                        raise IOCursorValidationError(
                            message=_(f'Cannot commit transactions to a non-existing ledger_xid {k}')
                        )

            elif isinstance(k, UUID):
                try:
                    self.commit_plan[self.ledger_map[k]] = txs
                except KeyError:
                    raise IOLibraryError(
                        message=_(f'Ledger UUID {k} not found.')
                    )

            elif isinstance(k, LedgerModel):
                self.commit_plan[k] = txs

            else:
                raise IOLibraryError('Unsupported ledger of type {x}'.format(x=type(k)))

        instructions = self.compile_instructions()
        account_codes = set(tx.account_code for tx in chain.from_iterable(tr for _, tr in instructions.items()))
        account_model_qs = self.resolve_account_model_qs(codes=account_codes)
        account_models = {
            acc.code: acc for acc in account_model_qs
        }

        for tx in chain.from_iterable(tr for _, tr in instructions.items()):
            try:
                tx.account_model = account_models[tx.account_code]
            except KeyError:
                raise IOCursorValidationError(
                    message=_(f'Account code {tx.account_code} not found. Is account available and not locked?')
                )

        results = dict()
        for ledger_model, tr_items in instructions.items():
            if ledger_model._state.adding:
                ledger_model.save()
            je_txs = [t.to_dict() for t in tr_items]

            # where the magic happens...
            je, txs_models = ledger_model.commit_txs(
                je_timestamp=je_timestamp if je_timestamp else get_localtime(),
                je_txs=je_txs,
                je_posted=post_journal_entries,
                je_desc=je_description,
                **kwargs
            )

            je.txs_models = txs_models

            results[ledger_model] = {
                'ledger_model': ledger_model,
                'journal_entry': je,
                'txs_models': txs_models,
                'instructions': tr_items,
                'account_model_qs': self.account_model_qs
            }

        self.__COMMITTED = True
        return results


class IOLibraryError(ValidationError):
    pass


class IOBluePrintValidationError(ValidationError):
    pass


class IOBluePrint:
    """
    This class documents instructions required to assemble and dispatch transactions into the ledger.

    Parameters
    ----------
    name: str, optional
        The human-readable name of the IOBluePrint instance.
    precision_decimals: int
        The number of decimals to use when balancing transactions. Defaults to 2.
    """

    def __init__(self, name: Optional[str] = None, precision_decimals: int = 2):
        self.name = name
        self.precision_decimals = precision_decimals
        self.registry = list()

    def get_name(self, entity_model: EntityModel) -> str:
        """
        Determines the name of the blueprint if none provided.

        Parameters
        ----------
        entity_model: EntityModel
            The EntityModel instance where the resulting blueprint transactions will be stored.

        Returns
        -------
        str
            The name of the blueprint.
        """
        if self.name is None:
            l_dt = get_localtime()
            return f'blueprint-{entity_model.slug}-{l_dt.strftime("%m%d%Y-%H%M%S")}'
        return self.name

    def _round_amount(self, amount: Decimal) -> Decimal:
        return round(amount, self.precision_decimals)

    def _amount(self, amount: Union[float, Decimal, int]) -> Decimal:
        if amount < 0:
            raise IOBluePrintValidationError(
                message='Amounts cannot be negative.'
            )

        if isinstance(amount, float):
            return self._round_amount(Decimal.from_float(amount))

        elif isinstance(amount, Decimal):
            return self._round_amount(amount)

        elif isinstance(amount, int):
            return Decimal(str(amount))

        raise IOBluePrintValidationError(
            message='Amounts must be float, Decimal or int.'
        )

    def credit(self, account_code: str, amount: Union[float, Decimal], description: str = None):
        """
        Registers a CREDIT to the specified account..

        Parameters
        ----------
        account_code: str
            The account code to use for the transaction.
        amount: float or Decimal
            The amount of the transaction.
        description: str
            Description of the transaction.
        """
        self.registry.append(
            TransactionInstructionItem(
                account_code=account_code,
                amount=self._amount(amount),
                tx_type=CREDIT,
                description=description
            ))

    def debit(self, account_code: str, amount: Union[float, Decimal], description: str = None):
        """
        Registers a DEBIT to the specified account.

        Parameters
        ----------
        account_code: str
            The account code to use for the transaction.
        amount: float or Decimal
            The amount of the transaction.
        description: str
            Description of the transaction.
        """
        self.registry.append(
            TransactionInstructionItem(
                account_code=account_code,
                amount=self._amount(amount),
                tx_type=DEBIT,
                description=description
            ))

    def commit(self,
               entity_model: EntityModel,
               user_model,
               ledger_model: Optional[Union[str, LedgerModel, UUID]] = None,
               je_timestamp: Optional[Union[datetime, date, str]] = None,
               post_new_ledgers: bool = False,
               post_journal_entries: bool = False,
               **kwargs) -> Dict:
        """
        Commits the blueprint transactions to the database.

        Parameters
        ----------
        entity_model: EntityModel
            The entity model instance where transactions will be committed.
        user_model: UserModel
            The user model instance executing transactions to check for permissions.
        ledger_model: Optional[Union[str, LedgerModel, UUID]]
            The ledger model instance identifier to be used for the transactions. If none, a new ledger will be created.
        je_timestamp: date or datetime or str, optional
            The date and/or time to be used for the transactions. If none, localtime will be used.
        post_new_ledgers: bool
            If True, newly created ledgers will be posted. Defaults to False.
        post_journal_entries: bool
            If True, newly created journal entries will be posted. Defaults to False.
        kwargs
            Keyword arguments passed to the IO Library.

        Returns
        -------
        Dict
            A dictionary containing the resulting models of the transactions.
        """
        blueprint_lib = IOLibrary(
            name=self.get_name(
                entity_model=entity_model
            ))

        cursor = blueprint_lib.get_cursor(
            entity_model=entity_model,
            user_model=user_model
        )

        cursor.blueprints[ledger_model].append(self)

        return cursor.commit(
            je_timestamp=je_timestamp,
            post_new_ledgers=post_new_ledgers,
            post_journal_entries=post_journal_entries,
            **kwargs
        )


class IOLibrary:
    """
    The IO Library is a centralized interface for documenting commonly used operations. The library will register and
    document the blueprints and their instructions so that they can be dispatched from anywhere in the application.

    Parameters
    ----------
    name: str
        The human-readable name of the library (i.e. PayRoll, Expenses, Rentals, etc...)
    """

    IO_CURSOR_CLASS = IOCursor

    def __init__(self, name: str):
        self.name = name
        self.registry: Dict[str, Callable] = {}

    def _check_func_name(self, name) -> bool:
        return name in self.registry

    def register(self, func: Callable):
        self.registry[func.__name__] = func

    def get_blueprint(self, name: str) -> Callable:
        """
        Retrieves a blueprint by name.

        Parameters
        ----------
        name: str
            The name of the blueprint to retrieve.

        Returns
        -------
        Callable
        """
        if not self._check_func_name(name):
            raise IOLibraryError(message=f'Function "{name}" is not registered in IO library {self.name}')
        return self.registry[name]

    def get_io_cursor_class(self):
        return self.IO_CURSOR_CLASS

    def get_cursor(
            self,
            entity_model: EntityModel,
            user_model,
            mode: IOCursorMode = IOCursorMode.PERMISSIVE,
            coa_model: Optional[Union[ChartOfAccountModel, UUID, str]] = None
    ) -> IOCursor:
        """
        Creates a cursor instance for associated with the library, entity and user model.

        Parameters
        ----------
        entity_model: EntityModel
            The entity model instance where transactions will be committed.
        user_model: UserModel
            The user model instance executing the transactions.
        coa_model: ChartOfAccountModel or UUID or str, optional
            The ChartOfAccountsModel instance or identifier used to determine the AccountModelQuerySet used for the transactions.
        mode: IOCursorMode
            The Mode of the cursor instance. Defaults to IOCursorMode.PERMISSIVE.

        Returns
        -------
        IOCursor
        """
        io_cursor_class = self.get_io_cursor_class()
        return io_cursor_class(
            io_library=self,
            entity_model=entity_model,
            user_model=user_model,
            coa_model=coa_model,
            mode=mode
        )



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from typing import List, Optional, Dict
from django_ledger.models.bank_account import BankAccountModel

from django.core.exceptions import ValidationError
from ofxtools import OFXTree
from ofxtools.models.bank import STMTRS
from ofxtools.models.ofx import OFX


class OFXImportValidationError(ValidationError):
    pass


class OFXFileManager:

    def __init__(self, ofx_file_or_path, parse_on_load: bool = True):
        self.FILE = ofx_file_or_path
        self.ofx_tree: OFXTree = OFXTree()
        self.ofx_data: OFX or None = None
        self.statements: List[STMTRS] or None = None
        self.NUMBER_OF_STATEMENTS: int = 0

        if parse_on_load:
            self.parse_ofx()

        if self.NUMBER_OF_STATEMENTS != 1:
            raise OFXImportValidationError('Only one account per OFX file is supported.')

        self.BANK_NAME = self.ofx_data.fi.org if hasattr(self.ofx_data.fi, 'org') else None
        self.FID = self.ofx_data.fi.fid if hasattr(self.ofx_data.fi, 'fid') else None
        self.ACCOUNT_DATA: Optional[Dict] = None

        self.get_account_data()
        # self.ACCOUNT_TXS = self.get_account_txs(account=self.ACCOUNTS[0]['account'].acctid)

    def parse_ofx(self):
        self.ofx_tree.parse(self.FILE)
        self.ofx_data = self.ofx_tree.convert()
        self.statements = self.ofx_data.statements
        self.NUMBER_OF_STATEMENTS = len(self.statements)

    def statement_attrs(self):
        return [a for a in dir(self.statements[0]) if a[0] != '_']

    def get_account_data(self):
        if self.ACCOUNT_DATA is None:
            self.ACCOUNT_DATA = [
                dict(
                    (attr, getattr(account, attr)) for attr in self.statement_attrs()
                ) | {
                    'bank': self.BANK_NAME,
                    'fid': self.FID
                } for account in self.statements
            ][0]
        return self.ACCOUNT_DATA

    def get_account_number(self):
        return self.get_account_data()['account'].acctid

    def get_routing_number(self):
        return self.get_account_data()['account'].bankid

    def get_ofx_account_type(self):
        """
        Gets the account type as defined in the OFX (Open Financial Exchange) specification.

        Returns:
            str: One of the following standardized account types:
                - 'CHECKING'   - Standard checking account
                - 'SAVINGS'    - Savings account
                - 'MONEYMRKT'  - Money Market account
                - 'CREDITLINE' - Credit line account
                - 'CD'         - Certificate of Deposit
        """
        acc_type = self.get_account_data()['account'].accttype

        if acc_type not in ['CHECKING', 'SAVINGS', 'MONEYMRKT', 'CREDITLINE', 'CD']:
            raise OFXImportValidationError(f'Account type "{acc_type}" is not supported.')
        return acc_type

    def get_account_type(self):
        return BankAccountModel.ACCOUNT_TYPE_OFX_MAPPING[self.get_ofx_account_type()]

    def get_account_txs(self):
        acc_statement = next(iter(
            st for st in self.ofx_data.statements if st.account.acctid == self.get_account_number()
        ))
        return acc_statement.banktranlist



from datetime import datetime, date
from typing import Optional, Dict, Union

from django_ledger.io.io_context import IODigestContextManager
from django_ledger.report.core import BaseReportSupport, PDFReportValidationError
from django_ledger.settings import DJANGO_LEDGER_CURRENCY_SYMBOL
from django_ledger.templatetags.django_ledger import currency_format


class CashFlowStatementReport(BaseReportSupport):

    def __init__(self, *args, io_digest: IODigestContextManager, report_subtitle: Optional[str] = None, **kwargs):

        if not io_digest.has_cash_flow_statement():
            raise PDFReportValidationError('IO Digest does not have income statement information.')
        super().__init__(*args, io_digest=io_digest, report_subtitle=report_subtitle, **kwargs)
        self.TABLE_HEADERS = {
            'empty_1': {
                'title': '',
                'spacing': 165,
                'align': 'L'
            },
            'total': {
                'title': f'Total ({DJANGO_LEDGER_CURRENCY_SYMBOL})',
                'spacing': 10,
                'align': 'R',
                'style': 'B'
            }
        }
        for k, th in self.TABLE_HEADERS.items():
            th['width'] = self.get_string_width(th['title']) + th['spacing']

    def get_report_data(self) -> Dict:
        return self.IO_DIGEST.get_cash_flow_statement_data()

    def get_report_name(self) -> str:
        return 'Cash Flow Statement'

    def print_amount(self, amt, zoom=0):
        self.set_x(178)
        self.set_font(
            family=self.FONT_FAMILY,
            size=self.FONT_SIZE + zoom
        )
        self.cell(
            w=20,
            h=5,
            markdown=True,
            align='R',
            txt=f'**{self.CURRENCY_SYMBOL}{currency_format(amt)}**'
        )
        self.set_default_font()

    def print_starting_net_income(self):
        net_income = self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_NET_INCOME']['balance']
        self.print_section_title(
            title=f'Net Income as of {self.IO_DIGEST.get_from_datetime(as_str=True)}',
            style='B',
            w=170)
        self.print_amount(amt=net_income, zoom=2)
        self.ln(8)

    def print_cash_from_operating(self):
        self.print_section_title(title='Cash from Operating Activities', style='U')
        self.ln(7)

        self.set_x(40)
        self.print_section_title(title='Non-cash Charges to Non-current Accounts', style='', zoom=2)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Depreciation & Amortization of Assets', style='', zoom=0, align='R', w=100)
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_OP_DEPRECIATION_AMORTIZATION'][
            'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Gain/Loss Sale of Assets', style='', zoom=0, align='R', w=100)
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_OP_INVESTMENT_GAINS'][
            'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.set_x(40)
        self.print_section_title(title='Non-cash Charges to Current Accounts', style='', zoom=2)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Accounts Receivable', style='', zoom=0, align='R', w=100)
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_OP_ACCOUNTS_RECEIVABLE'][
            'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Inventories', style='', zoom=0, align='R', w=100)
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_OP_INVENTORY'][
            'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Accounts Payable', style='', zoom=0, align='R', w=100)
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_OP_ACCOUNTS_PAYABLE'][
            'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Other Current Assets', style='', zoom=0, align='R', w=100)
        amt = \
            self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating']['GROUP_CFS_OP_OTHER_CURRENT_ASSETS_ADJUSTMENT'][
                'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.set_x(50)
        self.print_section_title(title='Other Current Liabilities', style='', zoom=0, align='R', w=100)
        amt = \
            self.IO_DIGEST.IO_DATA['cash_flow_statement']['operating'][
                'GROUP_CFS_OP_OTHER_CURRENT_LIABILITIES_ADJUSTMENT'][
                'balance']
        self.print_amount(amt=amt)
        self.ln(5)

        self.print_hline()
        self.print_section_title(
            title='Net Cash Provided by Operating Activities',
            zoom=1,
            w=100,
            style='B'
        )
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['net_cash_by_activity']['OPERATING']
        self.print_amount(amt=amt, zoom=1)
        self.ln(12)

    def print_cash_from_financing(self):
        self.print_section_title(title='Cash from Financing Activities', style='U')
        self.ln(7)

        cf_section = self.IO_DIGEST.IO_DATA['cash_flow_statement']['financing']
        for _, sec in cf_section.items():
            self.set_x(50)
            self.print_section_title(
                title=sec['description'],
                style='',
                zoom=0,
                align='R',
                w=100)
            self.print_amount(
                amt=sec['balance'])
            self.ln(5)

        self.print_hline()
        self.print_section_title(
            title='Net Cash Provided by Financing Activities',
            zoom=1,
            w=100,
            style='B'
        )
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['net_cash_by_activity']['FINANCING']
        self.print_amount(amt=amt, zoom=1)
        self.ln(12)

    def print_cash_from_investing(self):
        self.print_section_title(title='Cash from Investing Activities', style='U')
        self.ln(7)

        cf_section = self.IO_DIGEST.IO_DATA['cash_flow_statement']['investing']
        for _, sec in cf_section.items():
            self.set_x(50)
            self.print_section_title(
                title=sec['description'],
                style='',
                zoom=0,
                align='R',
                w=100)
            self.print_amount(
                amt=sec['balance'])
            self.ln(5)

        self.print_hline()
        self.print_section_title(
            title='Net Cash Provided by Investing Activities',
            zoom=1,
            w=100,
            style='B'
        )
        amt = self.IO_DIGEST.IO_DATA['cash_flow_statement']['net_cash_by_activity']['INVESTING']
        self.print_amount(amt=amt, zoom=1)
        self.ln(12)

    def print_net_cash_flow(self):
        self.print_section_title(title='Net Cash Flow', style='U')
        self.ln(7)

        self.set_x(50)
        self.print_section_title(
            title=f'Net Cash Flow from {self.IO_DIGEST.get_from_datetime(as_str=True)} '
                  f'through {self.IO_DIGEST.get_to_datetime(as_str=True)}',
            style='',
            zoom=0,
            align='R',
            w=100)
        self.print_amount(
            amt=self.IO_DIGEST.IO_DATA['cash_flow_statement']['net_cash'],
            zoom=3
        )
        self.ln(5)

    def get_pdf_filename(self,
                         from_dt: Optional[Union[datetime, date]] = None,
                         to_dt: Optional[Union[datetime, date]] = None,
                         dt_strfmt: str = '%Y%m%d'):
        if from_dt:
            from_dt = from_dt.strftime(dt_strfmt if dt_strfmt else self.IO_DIGEST.get_strftime_format())
        else:
            from_dt = self.IO_DIGEST.get_from_datetime(fmt=dt_strfmt, as_str=True)

        if to_dt:
            to_dt = to_dt.strftime(dt_strfmt if dt_strfmt else self.IO_DIGEST.get_strftime_format())
        else:
            to_dt = self.IO_DIGEST.get_to_datetime(fmt=dt_strfmt, as_str=True)
        f_name = f'{self.get_report_title()}_CashFlowStatement_{from_dt}_{to_dt}.pdf'
        return f_name

    def create_pdf_report(self):
        self.print_headers()
        self.print_starting_net_income()
        self.print_cash_from_operating()
        self.print_cash_from_financing()
        self.print_cash_from_investing()
        self.print_net_cash_flow()
